import string
from functools import lru_cache
from typing import List, Sequence, Optional, Tuple

from isla import language
from isla.helpers import srange
from isla.parser import PEGParser
from isla.type_defs import Grammar

from islearn.helpers import remove_spaces


def internet_checksum(
        _: Optional[Grammar],
        header: language.DerivationTree,
        checksum_tree: language.DerivationTree) -> language.SemPredEvalResult:
    if not header.is_complete():
        return language.SemPredEvalResult(None)

    checksum_grammar = {
        "<start>": ["<checksum>"],
        "<checksum>": ["<byte><byte>"],
        "<byte>": ["<zerof><zerof> "],
        "<zerof>": srange(string.digits + "ABCDEF")
    }

    checksum_tree_str = remove_spaces(checksum_tree)
    if not len(checksum_tree_str) % 2 == 0:
        return language.SemPredEvalResult(False)

    zeroes = "".join("0" for _ in range(len(checksum_tree_str)))
    if str(checksum_tree).endswith(" "):
        zeroes += " "

    zero_checksum = ("<checksum>", [(zeroes, [])])

    header_wo_checksum = header.replace_path(
        header.find_node(checksum_tree),
        language.DerivationTree.from_parse_tree(zero_checksum))

    header_bytes: Tuple[int] = tuple(hex_to_bytes(str(header_wo_checksum)))

    checksum_value = int_to_hex(compute_internet_checksum(header_bytes)).upper() + " "
    if len(checksum_value) < 6:
        assert len(checksum_value) == 3
        checksum_value = "00 " + checksum_value

    checksum_value_nospace = remove_spaces(checksum_value)
    if checksum_value_nospace == checksum_tree_str:
        return language.SemPredEvalResult(True)

    checksum_parser = PEGParser(checksum_grammar, start_symbol="<checksum>")
    new_checksum_tree = language.DerivationTree.from_parse_tree(
        list(checksum_parser.parse(checksum_value))[0])

    if str(new_checksum_tree) == str(checksum_tree):
        return language.SemPredEvalResult(True)

    return language.SemPredEvalResult({checksum_tree: new_checksum_tree})


INTERNET_CHECKSUM_PREDICATE = language.SemanticPredicate("internet_checksum", 2, internet_checksum, binds_tree=False)


@lru_cache(maxsize=128)
def compute_internet_checksum(inp_bytes: Tuple[int], length=16) -> int:
    if len(inp_bytes) % 2 != 0:
        inp_bytes = tuple(inp_bytes) + (0x00,)

    assert all(inp_byte <= 0xFF for inp_byte in inp_bytes)
    assert length < 8 or length % 8 == 0

    if length < 8:
        # This is only for testing purposes
        elements = list(inp_bytes)
    else:
        assert length == 16
        elements: List[int] = []
        for i in range(0, len(inp_bytes), 2):
            elements.append((inp_bytes[i] << 8) + inp_bytes[i + 1])

    carry_mask = 2 ** length
    no_carry_mask = 2 ** length - 1

    elements_sum = 0
    for element in elements:
        elements_sum += element
        elements_sum = ((elements_sum & no_carry_mask) + ((elements_sum & carry_mask) >> length))

    return elements_sum ^ no_carry_mask


def hex_to_bytes(hex_str: str) -> List[int]:
    return list(bytearray.fromhex(hex_str))


def hex_to_int(hex_str: str) -> int:
    length = len(remove_spaces(hex_str))
    if length % 2 != 0:
        hex_str = hex_str.rjust(length + 1, "0")
    return int.from_bytes(bytearray.fromhex(hex_str), byteorder="big")


def int_to_hex(number: int, add_spaces=True) -> str:
    result = list(hex(number)[2:])
    if len(result) % 2 != 0:
        result.insert(0, "0")
    if add_spaces:
        for i in reversed(range(len(result))):
            if i > 0 and i % 2 == 0:
                result.insert(i, " ")
    return "".join(result)


def bytes_to_hex(inp_bytes: Sequence[int], add_spaces=True) -> str:
    result = list(bytearray(inp_bytes).hex())
    if add_spaces:
        for i in reversed(range(len(result))):
            if i > 0 and i % 2 == 0:
                result.insert(i, " ")
    return "".join(result).upper()


def hex_to_dec(
        _: Optional[Grammar],
        hexadecimal: language.Variable | language.DerivationTree,
        decimal: language.Variable | language.DerivationTree) -> language.SemPredEvalResult:
    assert not isinstance(hexadecimal, language.Variable) or not isinstance(decimal, language.Variable)

    if (isinstance(hexadecimal, language.DerivationTree) and
            isinstance(decimal, language.DerivationTree) and
            not hexadecimal.is_complete() and
            not decimal.is_complete()):
        return language.SemPredEvalResult(None)

    bytes_grammar = {
        "<start>": ["<bytes>"],
        "<bytes>": ["<byte><bytes>", ""],
        "<byte>": ["<zerof><zerof> "],
        "<zerof>": srange(string.digits + "ABCDEF")
    }
    bytes_parser = PEGParser(bytes_grammar, start_symbol="bytes")

    if isinstance(hexadecimal, language.DerivationTree) and isinstance(decimal, language.DerivationTree):
        decimal_number = None
        hexadecimal_number = None
        if hexadecimal.is_complete():
            hexadecimal_number = hex_to_int(str(hexadecimal))
        if decimal.is_complete():
            decimal_number = int(str(decimal))

        if decimal_number == hexadecimal_number:
            return language.SemPredEvalResult(True)

        if hexadecimal.is_complete():
            return language.SemPredEvalResult({
                decimal: language.DerivationTree(
                    decimal.value,
                    (language.DerivationTree(str(hexadecimal_number), ()),))})
        else:
            return language.SemPredEvalResult({
                hexadecimal: language.DerivationTree(
                    hexadecimal.value,
                    (bytes_parser.parse(int_to_hex(decimal_number))[0]), )})

    if isinstance(hexadecimal, language.Variable) and isinstance(decimal, language.DerivationTree):
        if not decimal.is_complete():
            return language.SemPredEvalResult(None)

        return language.SemPredEvalResult({
            hexadecimal: language.DerivationTree(
                "<bytes>",
                (language.DerivationTree(
                    int_to_hex(int(str(decimal))),
                    ()),)
            )})

    if isinstance(decimal, language.Variable) and isinstance(hexadecimal, language.DerivationTree):
        if not hexadecimal.is_complete():
            return language.SemPredEvalResult(None)

        decimal_str = str(hex_to_int(str(hexadecimal)))

        return language.SemPredEvalResult({
            decimal: language.DerivationTree(
                "<decimal>",
                (language.DerivationTree(
                    decimal_str,
                    ()),)
            )})

    assert False


HEX_TO_DEC_PREDICATE = language.SemanticPredicate("hex_to_decimal", 2, hex_to_dec, binds_tree=False)



# Generated from IslaLanguage.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .IslaLanguageParser import IslaLanguageParser
else:
    from IslaLanguageParser import IslaLanguageParser

# This class defines a complete listener for a parse tree produced by IslaLanguageParser.
class IslaLanguageListener(ParseTreeListener):

    # Enter a parse tree produced by IslaLanguageParser#start.
    def enterStart(self, ctx:IslaLanguageParser.StartContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#start.
    def exitStart(self, ctx:IslaLanguageParser.StartContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#constDecl.
    def enterConstDecl(self, ctx:IslaLanguageParser.ConstDeclContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#constDecl.
    def exitConstDecl(self, ctx:IslaLanguageParser.ConstDeclContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ExistsMexpr.
    def enterExistsMexpr(self, ctx:IslaLanguageParser.ExistsMexprContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ExistsMexpr.
    def exitExistsMexpr(self, ctx:IslaLanguageParser.ExistsMexprContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Negation.
    def enterNegation(self, ctx:IslaLanguageParser.NegationContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Negation.
    def exitNegation(self, ctx:IslaLanguageParser.NegationContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Implication.
    def enterImplication(self, ctx:IslaLanguageParser.ImplicationContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Implication.
    def exitImplication(self, ctx:IslaLanguageParser.ImplicationContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ForallMexpr.
    def enterForallMexpr(self, ctx:IslaLanguageParser.ForallMexprContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ForallMexpr.
    def exitForallMexpr(self, ctx:IslaLanguageParser.ForallMexprContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ExistsInt.
    def enterExistsInt(self, ctx:IslaLanguageParser.ExistsIntContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ExistsInt.
    def exitExistsInt(self, ctx:IslaLanguageParser.ExistsIntContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Disjunction.
    def enterDisjunction(self, ctx:IslaLanguageParser.DisjunctionContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Disjunction.
    def exitDisjunction(self, ctx:IslaLanguageParser.DisjunctionContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#PredicateAtom.
    def enterPredicateAtom(self, ctx:IslaLanguageParser.PredicateAtomContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#PredicateAtom.
    def exitPredicateAtom(self, ctx:IslaLanguageParser.PredicateAtomContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SMTFormula.
    def enterSMTFormula(self, ctx:IslaLanguageParser.SMTFormulaContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SMTFormula.
    def exitSMTFormula(self, ctx:IslaLanguageParser.SMTFormulaContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Equivalence.
    def enterEquivalence(self, ctx:IslaLanguageParser.EquivalenceContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Equivalence.
    def exitEquivalence(self, ctx:IslaLanguageParser.EquivalenceContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Exists.
    def enterExists(self, ctx:IslaLanguageParser.ExistsContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Exists.
    def exitExists(self, ctx:IslaLanguageParser.ExistsContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Conjunction.
    def enterConjunction(self, ctx:IslaLanguageParser.ConjunctionContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Conjunction.
    def exitConjunction(self, ctx:IslaLanguageParser.ConjunctionContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ParFormula.
    def enterParFormula(self, ctx:IslaLanguageParser.ParFormulaContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ParFormula.
    def exitParFormula(self, ctx:IslaLanguageParser.ParFormulaContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#Forall.
    def enterForall(self, ctx:IslaLanguageParser.ForallContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#Forall.
    def exitForall(self, ctx:IslaLanguageParser.ForallContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ExclusiveOr.
    def enterExclusiveOr(self, ctx:IslaLanguageParser.ExclusiveOrContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ExclusiveOr.
    def exitExclusiveOr(self, ctx:IslaLanguageParser.ExclusiveOrContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#ForallInt.
    def enterForallInt(self, ctx:IslaLanguageParser.ForallIntContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#ForallInt.
    def exitForallInt(self, ctx:IslaLanguageParser.ForallIntContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprNonterminalStringPh.
    def enterSexprNonterminalStringPh(self, ctx:IslaLanguageParser.SexprNonterminalStringPhContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprNonterminalStringPh.
    def exitSexprNonterminalStringPh(self, ctx:IslaLanguageParser.SexprNonterminalStringPhContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprInfixReStr.
    def enterSexprInfixReStr(self, ctx:IslaLanguageParser.SexprInfixReStrContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprInfixReStr.
    def exitSexprInfixReStr(self, ctx:IslaLanguageParser.SexprInfixReStrContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprNum.
    def enterSexprNum(self, ctx:IslaLanguageParser.SexprNumContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprNum.
    def exitSexprNum(self, ctx:IslaLanguageParser.SexprNumContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprXPathExpr.
    def enterSexprXPathExpr(self, ctx:IslaLanguageParser.SexprXPathExprContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprXPathExpr.
    def exitSexprXPathExpr(self, ctx:IslaLanguageParser.SexprXPathExprContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprOp.
    def enterSexprOp(self, ctx:IslaLanguageParser.SexprOpContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprOp.
    def exitSexprOp(self, ctx:IslaLanguageParser.SexprOpContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprInfixPlusMinus.
    def enterSexprInfixPlusMinus(self, ctx:IslaLanguageParser.SexprInfixPlusMinusContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprInfixPlusMinus.
    def exitSexprInfixPlusMinus(self, ctx:IslaLanguageParser.SexprInfixPlusMinusContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprTrue.
    def enterSexprTrue(self, ctx:IslaLanguageParser.SexprTrueContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprTrue.
    def exitSexprTrue(self, ctx:IslaLanguageParser.SexprTrueContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprFreeId.
    def enterSexprFreeId(self, ctx:IslaLanguageParser.SexprFreeIdContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprFreeId.
    def exitSexprFreeId(self, ctx:IslaLanguageParser.SexprFreeIdContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprStringPh.
    def enterSexprStringPh(self, ctx:IslaLanguageParser.SexprStringPhContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprStringPh.
    def exitSexprStringPh(self, ctx:IslaLanguageParser.SexprStringPhContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SepxrApp.
    def enterSepxrApp(self, ctx:IslaLanguageParser.SepxrAppContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SepxrApp.
    def exitSepxrApp(self, ctx:IslaLanguageParser.SepxrAppContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprInfixMulDiv.
    def enterSexprInfixMulDiv(self, ctx:IslaLanguageParser.SexprInfixMulDivContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprInfixMulDiv.
    def exitSexprInfixMulDiv(self, ctx:IslaLanguageParser.SexprInfixMulDivContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprId.
    def enterSexprId(self, ctx:IslaLanguageParser.SexprIdContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprId.
    def exitSexprId(self, ctx:IslaLanguageParser.SexprIdContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprPrefix.
    def enterSexprPrefix(self, ctx:IslaLanguageParser.SexprPrefixContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprPrefix.
    def exitSexprPrefix(self, ctx:IslaLanguageParser.SexprPrefixContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprStr.
    def enterSexprStr(self, ctx:IslaLanguageParser.SexprStrContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprStr.
    def exitSexprStr(self, ctx:IslaLanguageParser.SexprStrContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprInfixEq.
    def enterSexprInfixEq(self, ctx:IslaLanguageParser.SexprInfixEqContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprInfixEq.
    def exitSexprInfixEq(self, ctx:IslaLanguageParser.SexprInfixEqContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprDisjStringsPh.
    def enterSexprDisjStringsPh(self, ctx:IslaLanguageParser.SexprDisjStringsPhContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprDisjStringsPh.
    def exitSexprDisjStringsPh(self, ctx:IslaLanguageParser.SexprDisjStringsPhContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#SexprFalse.
    def enterSexprFalse(self, ctx:IslaLanguageParser.SexprFalseContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#SexprFalse.
    def exitSexprFalse(self, ctx:IslaLanguageParser.SexprFalseContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#predicateArg.
    def enterPredicateArg(self, ctx:IslaLanguageParser.PredicateArgContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#predicateArg.
    def exitPredicateArg(self, ctx:IslaLanguageParser.PredicateArgContext):
        pass


    # Enter a parse tree produced by IslaLanguageParser#smt_binary_op.
    def enterSmt_binary_op(self, ctx:IslaLanguageParser.Smt_binary_opContext):
        pass

    # Exit a parse tree produced by IslaLanguageParser#smt_binary_op.
    def exitSmt_binary_op(self, ctx:IslaLanguageParser.Smt_binary_opContext):
        pass



del IslaLanguageParser


import copy
import functools
import inspect
import itertools
import logging
import os.path
import pkgutil
import string
from abc import ABC
from functools import lru_cache
from typing import List, Tuple, Set, Dict, Optional, cast, Callable, Iterable, Sequence

import datrie
import isla.fuzzer
import toml
import z3
from grammar_graph import gg
from isla import language, isla_predicates
from isla.evaluator import evaluate, matches_for_quantified_formula
from isla.helpers import RE_NONTERMINAL, weighted_geometric_mean, \
    is_nonterminal, dict_of_lists_to_list_of_dicts, canonical
from isla.isla_predicates import reachable
from isla.language import set_smt_auto_eval, ensure_unique_bound_variables
from isla.solver import ISLaSolver
from isla.three_valued_truth import ThreeValuedTruth
from isla.trie import path_to_trie_key
from isla.type_defs import Grammar, ParseTree, Path
from isla.z3_helpers import z3_subst, evaluate_z3_expression, is_valid, \
    DomainError
from pathos import multiprocessing as pmp

from islearn.helpers import connected_chains, transitive_closure, tree_in, \
    is_int, is_float, e_assert
from islearn.language import NonterminalPlaceholderVariable, PlaceholderVariable, \
    NonterminalStringPlaceholderVariable, parse_abstract_isla, StringPlaceholderVariable, \
    AbstractISLaUnparser, MexprPlaceholderVariable, AbstractBindExpression, DisjunctiveStringsPlaceholderVariable, \
    StringPlaceholderVariableTypes
from islearn.mutation import MutationFuzzer
from islearn.parse_tree_utils import replace_path, expand_tree, tree_leaves, \
    get_subtree, tree_paths, trie_from_parse_tree, next_trie_key, tree_from_paths, Tree, get_subtrie
from islearn.reducer import InputReducer

STANDARD_PATTERNS_REPO = "patterns.toml"
logger = logging.getLogger("learner")


class TruthTableRow:
    def __init__(
            self,
            formula: language.Formula,
            inputs: Sequence[language.DerivationTree] = (),
            eval_results: Sequence[bool] = ()):
        self.formula = formula
        self.inputs = list(inputs)
        self.eval_results: List[bool] = list(eval_results)

    def evaluate(
            self,
            graph: gg.GrammarGraph,
            columns_parallel: bool = False,
            lazy: bool = False,
            result_threshold: float = .9) -> 'TruthTableRow':
        """If lazy is True, then the evaluation stops as soon as result_threshold can no longer be
        reached. E.g., if result_threshold is .9 and there are 100 inputs, then after more than
        10 negative results, 90% positive results is no longer possible."""

        if columns_parallel:
            with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
                iterator = pool.imap(
                    lambda inp: evaluate(self.formula, inp, graph.grammar, graph=graph).is_true(),
                    self.inputs)

                self.eval_results = []
                negative_results = 0
                for eval_result in iterator:
                    if lazy and negative_results > len(self.inputs) * (1 - result_threshold):
                        self.eval_results += [False for _ in range(len(self.inputs) - len(self.eval_results))]
                        break

                    if not eval_result:
                        negative_results += 1
                    self.eval_results.append(eval_result)
        else:
            self.eval_results = []
            negative_results = 0
            for inp in self.inputs:
                if lazy and negative_results > len(self.inputs) * (1 - result_threshold):
                    self.eval_results += [False for _ in range(len(self.inputs) - len(self.eval_results))]
                    break

                eval_result = evaluate(self.formula, inp, graph.grammar, graph=graph).is_true()
                if not eval_result:
                    negative_results += 1
                self.eval_results.append(eval_result)

        return self

    def eval_result(self) -> float:
        assert len(self.inputs) > 0
        assert len(self.eval_results) == len(self.inputs)
        assert all(isinstance(entry, bool) for entry in self.eval_results)
        return sum(int(entry) for entry in self.eval_results) / len(self.eval_results)

    def __repr__(self):
        return f"TruthTableRow({repr(self.formula)}, {repr(self.inputs)}, {repr(self.eval_results)})"

    def __str__(self):
        return f"{self.formula}: {', '.join(map(str, self.eval_results))}"

    def __eq__(self, other):
        return (isinstance(other, TruthTableRow) and
                self.formula == other.formula)

    def __len__(self):
        return len(self.eval_results)

    def __hash__(self):
        return hash(self.formula)

    def __neg__(self):
        return TruthTableRow(
            -self.formula,
            self.inputs,
            [not eval_result for eval_result in self.eval_results]
        )

    def __and__(self, other: 'TruthTableRow') -> 'TruthTableRow':
        assert len(self.inputs) == len(other.inputs)
        assert len(self.eval_results) == len(other.eval_results)
        return TruthTableRow(
            self.formula & other.formula,
            self.inputs,
            [a and b for a, b in zip(self.eval_results, other.eval_results)]
        )

    def __or__(self, other: 'TruthTableRow') -> 'TruthTableRow':
        assert len(self.inputs) == len(other.inputs)
        assert len(self.eval_results) == len(other.eval_results)
        return TruthTableRow(
            self.formula | other.formula,
            self.inputs,
            [a or b for a, b in zip(self.eval_results, other.eval_results)]
        )


class TruthTable:
    def __init__(self, rows: Iterable[TruthTableRow] = ()):
        self.__row_hashes = set([])
        self.__rows = []
        for row in rows:
            row_hash = hash(row)
            if row_hash not in self.__row_hashes:
                self.__row_hashes.add(row_hash)
                self.__rows.append(row)

    def __repr__(self):
        return f"TruthTable({repr(self.__rows)})"

    def __str__(self):
        return "\n".join(map(str, self.__rows))

    def __getitem__(self, item: int | language.Formula) -> TruthTableRow:
        if isinstance(item, int):
            return self.__rows[item]

        assert isinstance(item, language.Formula)

        try:
            return next(row for row in self.__rows if row.formula == item)
        except StopIteration:
            raise KeyError(item)

    def __len__(self):
        return len(self.__rows)

    def __iter__(self):
        return iter(self.__rows)

    def append(self, row: TruthTableRow):
        row_hash = hash(row)
        if row_hash not in self.__row_hashes:
            self.__row_hashes.add(row_hash)
            self.__rows.append(row)

    def remove(self, row: TruthTableRow):
        self.__rows.remove(row)
        self.__row_hashes.remove(hash(row))

    def __add__(self, other: 'TruthTable') -> 'TruthTable':
        return TruthTable(self.__rows + other.__rows)

    def __iadd__(self, other: 'TruthTable') -> 'TruthTable':
        for row in other.__rows:
            self.append(row)

        return self

    def evaluate(
            self,
            graph: gg.GrammarGraph,
            columns_parallel: bool = False,
            rows_parallel: bool = False,
            lazy: bool = False,
            result_threshold: float = .9) -> 'TruthTable':
        """If lazy is True, then column evaluation stops as soon as result_threshold can no longer be
        reached. E.g., if result_threshold is .9 and there are 100 inputs, then after more than
        10 negative results, 90% positive results is no longer possible."""

        assert not columns_parallel or not rows_parallel

        if rows_parallel:
            with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
                self.__rows = set(pool.map(
                    lambda row: row.evaluate(graph, columns_parallel, lazy=lazy, result_threshold=result_threshold),
                    self.__rows
                ))
        else:
            for row in self.__rows:
                row.evaluate(graph, columns_parallel, lazy=lazy, result_threshold=result_threshold)

        return self


class InvariantLearner:
    def __init__(
            self,
            grammar: Grammar,
            prop: Optional[Callable[[language.DerivationTree], bool]] = None,
            positive_examples: Optional[Iterable[language.DerivationTree]] = None,
            negative_examples: Optional[Iterable[language.DerivationTree]] = None,
            patterns: Optional[List[language.Formula | str]] = None,
            pattern_file: Optional[str] = None,
            activated_patterns: Optional[Iterable[str]] = None,
            deactivated_patterns: Optional[Iterable[str]] = None,
            k: int = 3,
            target_number_positive_samples: int = 10,
            target_number_negative_samples: int = 10,
            target_number_positive_samples_for_learning: int = 10,
            mexpr_expansion_limit: int = 1,
            max_nonterminals_in_mexpr: Optional[int] = None,
            min_recall: float = .9,
            min_specificity: float = .6,
            max_disjunction_size: int = 1,
            max_conjunction_size: int = 2,
            exclude_nonterminals: Optional[Iterable[str]] = None,
            include_negations_in_disjunctions: bool = False,
            reduce_inputs_for_learning: bool = True,
            reduce_all_inputs: bool = False,
            generate_new_learning_samples: bool = True,
            do_generate_more_inputs: bool = True,
            filter_inputs_for_learning_by_kpaths: bool = True,
    ):
        # We add extended caching certain, crucial functions.
        isla.helpers.evaluate_z3_expression = lru_cache(maxsize=None)(
            inspect.unwrap(evaluate_z3_expression))
        isla.language.DerivationTree.__str__ = lru_cache(maxsize=None)(
            inspect.unwrap(isla.language.DerivationTree.__str__))
        isla.language.DerivationTree.paths = lru_cache(maxsize=128)(
            inspect.unwrap(isla.language.DerivationTree.paths))
        isla.language.DerivationTree.__hash__ = lambda tree: tree.id
        isla.isla_predicates.is_nth = lru_cache(maxsize=128)(
            inspect.unwrap(isla.isla_predicates.is_nth))

        self.grammar = grammar
        self.canonical_grammar = canonical(grammar)
        self.graph = gg.GrammarGraph.from_grammar(grammar)
        self.prop = prop
        self.k = k
        self.mexpr_expansion_limit = mexpr_expansion_limit
        self.max_nonterminals_in_mexpr = max_nonterminals_in_mexpr
        self.min_recall = min_recall
        self.min_specificity = min_specificity
        self.max_disjunction_size = max_disjunction_size
        self.max_conjunction_size = max_conjunction_size
        self.exclude_nonterminals = exclude_nonterminals or set([])
        self.include_negations_in_disjunctions = include_negations_in_disjunctions
        self.reduce_inputs_for_learning = reduce_inputs_for_learning
        self.reduce_all_inputs = reduce_all_inputs
        self.generate_new_learning_samples = generate_new_learning_samples
        self.do_generate_more_inputs = do_generate_more_inputs
        self.filter_inputs_for_learning_by_kpaths = filter_inputs_for_learning_by_kpaths

        self.positive_examples: List[language.DerivationTree] = list(set(positive_examples or []))
        self.original_positive_examples: List[language.DerivationTree] = list(self.positive_examples)
        self.negative_examples: List[language.DerivationTree] = list(set(negative_examples or []))
        self.positive_examples_for_learning: List[language.DerivationTree] = []

        self.target_number_positive_samples = target_number_positive_samples
        self.target_number_negative_samples = target_number_negative_samples
        self.target_number_positive_samples_for_learning = target_number_positive_samples_for_learning
        assert target_number_positive_samples >= target_number_positive_samples_for_learning

        assert not prop or all(prop(example) for example in self.positive_examples)
        assert not prop or all(not prop(example) for example in self.negative_examples)

        # Also consider inverted patterns?
        assert not activated_patterns or not deactivated_patterns
        if not patterns:
            pattern_repo = patterns_from_file(pattern_file or STANDARD_PATTERNS_REPO)
            if activated_patterns:
                self.patterns = [pattern for name in activated_patterns for pattern in pattern_repo[name]]
            else:
                self.patterns = list(pattern_repo.get_all(but=deactivated_patterns or []))
        else:
            self.patterns = [
                pattern if isinstance(pattern, language.Formula)
                else parse_abstract_isla(pattern, grammar)
                for pattern in patterns]

    def learn_invariants(self, ensure_unique_var_names: bool = True) -> Dict[language.Formula, Tuple[float, float]]:
        if self.prop and self.do_generate_more_inputs:
            self._generate_more_inputs()
            assert len(self.positive_examples) > 0, "Cannot learn without any positive examples!"
            assert all(self.prop(positive_example) for positive_example in self.positive_examples)
            assert all(not self.prop(negative_example) for negative_example in self.negative_examples)

        if self.reduce_all_inputs and self.prop is not None:
            logger.info(
                "Reducing %d positive samples w.r.t. property and k=%d.",
                len(self.positive_examples),
                self.k)
            reducer = InputReducer(self.grammar, self.prop, self.k)
            self.positive_examples = [
                reducer.reduce_by_smallest_subtree_replacement(inp)
                for inp in self.positive_examples]

            logger.info(
                "Reducing %d negative samples w.r.t. property and k=%d.",
                len(self.negative_examples),
                self.k)
            reducer = InputReducer(self.grammar, lambda t: not self.prop(t), self.k)
            self.negative_examples = [
                reducer.reduce_by_smallest_subtree_replacement(inp)
                for inp in self.negative_examples]

        if self.generate_new_learning_samples or not self.original_positive_examples:
            self.positive_examples_for_learning = \
                self._sort_inputs(
                    self.positive_examples,
                    self.filter_inputs_for_learning_by_kpaths,
                    more_paths_weight=1.7,
                    smaller_inputs_weight=1.0)[:self.target_number_positive_samples_for_learning]
        else:
            self.positive_examples_for_learning = \
                self._sort_inputs(
                    self.original_positive_examples,
                    self.filter_inputs_for_learning_by_kpaths,
                    more_paths_weight=1.7,
                    smaller_inputs_weight=1.0)[:self.target_number_positive_samples_for_learning]

        logger.info(
            "Keeping %d positive examples for candidate generation.",
            len(self.positive_examples_for_learning)
        )

        if ((not self.reduce_all_inputs or not self.generate_new_learning_samples)
                and self.reduce_inputs_for_learning
                and self.prop is not None):
            logger.info(
                "Reducing %d inputs for learning w.r.t. property and k=%d.",
                len(self.positive_examples_for_learning),
                self.k
            )
            reducer = InputReducer(self.grammar, self.prop, self.k)
            self.positive_examples_for_learning = [
                reducer.reduce_by_smallest_subtree_replacement(inp)
                for inp in self.positive_examples_for_learning]

        logger.debug(
            "Examples for learning:\n%s",
            "\n".join(map(str, self.positive_examples_for_learning)))

        candidates = self.generate_candidates(self.patterns, self.positive_examples_for_learning)
        logger.info("Found %d invariant candidates.", len(candidates))

        logger.debug(
            "Candidates:\n%s",
            "\n\n".join([language.ISLaUnparser(candidate).unparse() for candidate in candidates]))

        logger.info("Filtering invariants.")

        # Only consider *real* invariants

        # NOTE: Disabled parallel evaluation for now. In certain cases, this renders
        #       the filtering process *much* slower, or gives rise to stack overflows
        #       (e.g., "test_learn_from_islearn_patterns_file" example).
        recall_truth_table = TruthTable([
            TruthTableRow(inv, self.positive_examples)
            for inv in candidates
        ]).evaluate(
            self.graph,
            # rows_parallel=True,
            lazy=self.max_disjunction_size < 2,
            result_threshold=self.min_recall
        )

        if self.max_disjunction_size < 2:
            for row in recall_truth_table:
                if row.eval_result() < self.min_recall:
                    recall_truth_table.remove(row)

        precision_truth_table = None
        if self.negative_examples:
            logger.info("Evaluating precision.")
            logger.debug("Negative samples:\n" + "\n-----------\n".join(map(str, self.negative_examples)))

            precision_truth_table = TruthTable([
                TruthTableRow(row.formula, self.negative_examples)
                for row in recall_truth_table
            ]).evaluate(
                self.graph,
                # rows_parallel=True
            )

            assert len(precision_truth_table) == len(recall_truth_table)

        assert not self.negative_examples or precision_truth_table is not None

        invariants = {
            row.formula for row in recall_truth_table
            if row.eval_result() >= self.min_recall
        }

        logger.info(
            "%d invariants with recall >= %d%% remain after filtering.",
            len(invariants),
            int(self.min_recall * 100))

        logger.debug("Invariants:\n%s", "\n\n".join(map(lambda f: language.ISLaUnparser(f).unparse(), invariants)))

        if self.max_disjunction_size > 1:
            logger.info("Calculating recall of Boolean combinations.")

            disjunctive_recall_truthtable = copy.deepcopy(recall_truth_table)
            assert precision_truth_table is None or len(disjunctive_recall_truthtable) == len(precision_truth_table)

            for level in range(2, self.max_disjunction_size + 1):
                assert precision_truth_table is None or len(disjunctive_recall_truthtable) == len(precision_truth_table)
                logger.debug(f"Disjunction size: {level}")

                for rows_with_indices in itertools.combinations(enumerate(recall_truth_table), level):
                    assert (precision_truth_table is None or
                            len(disjunctive_recall_truthtable) == len(precision_truth_table))
                    assert (precision_truth_table is None or
                            all(rwi[1].formula == precision_truth_table[rwi[0]].formula for rwi in rows_with_indices))

                    max_num_negations = level // 2 if self.include_negations_in_disjunctions else 0
                    for formulas_to_negate in (
                            t for t in itertools.product(*[[0, 1] for _ in range(3)]) if sum(t) <= max_num_negations):

                        # To ensure that only "meaningful" properties are negated, the un-negated properties
                        # should hold for at least 20% of all inputs; but at most for 80%, since otherwise,
                        # the negation is overly specific.
                        negated_rows: List[Tuple[int, TruthTableRow]] = list(
                            itertools.compress(rows_with_indices, formulas_to_negate))
                        if any(.8 < negated_row.eval_result() < .2 for _, negated_row in negated_rows):
                            continue

                        recall_table_rows = [
                            -row if bool(negate) else row
                            for negate, (_, row) in zip(formulas_to_negate, rows_with_indices)]

                        # Compute recall of disjunction, add if above threshold and
                        # an improvement over all participants of the disjunction
                        disjunction = functools.reduce(TruthTableRow.__or__, recall_table_rows)
                        new_recall = disjunction.eval_result()
                        if (new_recall < self.min_recall or
                                not all(new_recall > row.eval_result() for row in recall_table_rows)):
                            continue

                        disjunctive_recall_truthtable.append(disjunction)

                        if precision_truth_table is not None:
                            # Also add disjunction to the precision truth table. Saves us a couple of evaluations.
                            precision_table_rows = [
                                -precision_truth_table[idx] if bool(negate) else precision_truth_table[idx]
                                for negate, (idx, _) in zip(formulas_to_negate, rows_with_indices)]
                            disjunction = functools.reduce(TruthTableRow.__or__, precision_table_rows)
                            precision_truth_table.append(disjunction)

            recall_truth_table = disjunctive_recall_truthtable

            invariants = {
                row.formula for row in recall_truth_table
                if row.eval_result() >= self.min_recall
            }

            logger.info(
                "%d invariants with recall >= %d%% remain after building Boolean combinations.",
                len(invariants),
                int(self.min_recall * 100))

            # logger.debug(
            #   "Invariants:\n%s", "\n\n".join(map(lambda f: language.ISLaUnparser(f).unparse(), invariants)))

        # assert all(evaluate(row.formula, inp, self.grammar).is_true()
        #            for inp in self.positive_examples
        #            for row in recall_truth_table
        #            if row.eval_result() == 1)

        if not self.negative_examples:
            # TODO: Enforce unique names, sort
            if ensure_unique_var_names:
                invariants = sorted(list(map(ensure_unique_bound_variables, invariants)), key=lambda inv: len(inv))
            else:
                invariants = sorted(list(invariants), key=lambda inv: len(inv))
            return {
                row.formula: (1.0, row.eval_result())
                for row in recall_truth_table
                if row.eval_result() >= self.min_recall
            }

        indices_to_remove = list(reversed([
            idx for idx, row in enumerate(recall_truth_table)
            if row.eval_result() < self.min_recall]))
        assert sorted(indices_to_remove, reverse=True) == indices_to_remove
        for idx in indices_to_remove:
            recall_truth_table.remove(recall_truth_table[idx])
            precision_truth_table.remove(precision_truth_table[idx])

        logger.info("Calculating precision of Boolean combinations.")
        conjunctive_precision_truthtable = copy.copy(precision_truth_table)
        for level in range(2, self.max_conjunction_size + 1):
            logger.debug(f"Conjunction size: {level}")
            assert len(recall_truth_table) == len(conjunctive_precision_truthtable)
            for rows_with_indices in itertools.combinations(enumerate(precision_truth_table), level):
                precision_table_rows = [row for (_, row) in rows_with_indices]

                # Only consider combinations where all rows meet minimum recall requirement.
                # Recall doesn't get better by forming conjunctions!
                if any(recall_truth_table[idx].eval_result() < self.min_recall
                       for idx, _ in rows_with_indices):
                    continue

                # Compute precision of conjunction, add if above threshold and
                # an improvement over all participants of the conjunction
                conjunction = functools.reduce(TruthTableRow.__and__, precision_table_rows)
                new_precision = 1 - conjunction.eval_result()
                if (new_precision < self.min_specificity or
                        not all(new_precision > 1 - row.eval_result() for row in precision_table_rows)):
                    continue

                conjunctive_precision_truthtable.append(conjunction)

                recall_table_rows = [recall_truth_table[idx] for idx, _ in rows_with_indices]
                conjunction = functools.reduce(TruthTableRow.__and__, recall_table_rows)
                recall_truth_table.append(conjunction)

        precision_truth_table = conjunctive_precision_truthtable

        # assert all(evaluate(row.formula, inp, self.grammar).is_false()
        #            for inp in self.negative_examples
        #            for row in precision_truth_table
        #            if row.eval_result() == 0)

        result: Dict[language.Formula, Tuple[float, float]] = {
            precision_row.formula if not ensure_unique_var_names
            else language.ensure_unique_bound_variables(precision_row.formula):
                (1 - precision_row.eval_result(), recall_truth_table[idx].eval_result())
            for idx, precision_row in enumerate(precision_truth_table)
            if (1 - precision_row.eval_result() >= self.min_specificity and
                recall_truth_table[idx].eval_result() >= self.min_recall)
        }

        logger.info(
            "Found %d invariants with precision >= %d%%.",
            len([p for p in result.values() if p[0] >= self.min_specificity]),
            int(self.min_specificity * 100),
        )

        # TODO: Sort within same recall / specificity values: Fewer disjunctions,
        #       more common String constants... To optimize specificity further.

        return dict(
            cast(List[Tuple[language.Formula, Tuple[float, float]]],
                 sorted(result.items(),
                        key=lambda p: (p[1], -len(p[0])),
                        reverse=True)))

    def generate_candidates(
            self,
            patterns: Iterable[language.Formula | str],
            inputs: Iterable[language.DerivationTree]) -> Set[language.Formula]:
        input_reachability_relation = create_input_reachability_relation(inputs)

        logger.debug("Computed input reachability relation of size %d", len(input_reachability_relation))

        filters: List[PatternInstantiationFilter] = [
            NonterminalStringInCountPredicatesFilter(self.graph, input_reachability_relation),
        ]

        patterns = [
            parse_abstract_isla(pattern, self.grammar) if isinstance(pattern, str)
            else pattern
            for pattern in patterns]

        tries: List[datrie.Trie] = [inp.trie().trie for inp in inputs]

        result: Set[language.Formula] = set([])

        for pattern in patterns:
            logger.debug("Instantiating pattern\n%s", AbstractISLaUnparser(pattern).unparse())
            set_smt_auto_eval(pattern, False)

            # Instantiate various placeholder variables:
            # 1. Nonterminal placeholders
            pattern_insts_without_nonterminal_placeholders = self._instantiate_nonterminal_placeholders(
                pattern, input_reachability_relation)

            logger.debug("Found %d instantiations of pattern meeting quantifier requirements",
                         len(pattern_insts_without_nonterminal_placeholders))

            # NOTE: At this point, filtering is not useful. It is cheaper to first instantiate
            #       match expressions (if any), which reduces the search space, and to filter then.
            # pattern_insts_without_nonterminal_placeholders = self._filter_partial_instantiations(
            #     pattern_insts_without_nonterminal_placeholders, tries)
            # logger.debug("%d instantiations remain after filtering",
            #              len(pattern_insts_without_nonterminal_placeholders))

            # 2. Match expression placeholders
            pattern_insts_without_mexpr_placeholders = self._instantiate_mexpr_placeholders(
                pattern_insts_without_nonterminal_placeholders)

            logger.debug("Found %d instantiations of pattern after instantiating match expression placeholders",
                         len(pattern_insts_without_mexpr_placeholders))

            pattern_insts_without_mexpr_placeholders = self._filter_partial_instantiations(
                pattern_insts_without_mexpr_placeholders, tries)
            logger.debug("%d instantiations remain after filtering",
                         len(pattern_insts_without_mexpr_placeholders))

            # 3. Special string placeholders in predicates.
            #    This comprises, e.g., `nth(<STRING>, elem, container`.
            pattern_insts_without_special_string_placeholders = \
                self.__instantiate_special_predicate_string_placeholders(
                    pattern_insts_without_mexpr_placeholders, tries)

            logger.debug("Found %d instantiations of pattern after instantiating special predicate string placeholders",
                         len(pattern_insts_without_special_string_placeholders))

            if pattern_insts_without_special_string_placeholders != pattern_insts_without_mexpr_placeholders:
                pattern_insts_without_special_string_placeholders = self._filter_partial_instantiations(
                    pattern_insts_without_special_string_placeholders, tries)
                logger.debug("%d instantiations remain after filtering",
                             len(pattern_insts_without_special_string_placeholders))

            # 4. Nonterminal-String placeholders
            pattern_insts_without_nonterminal_string_placeholders = self._instantiate_nonterminal_string_placeholders(
                pattern_insts_without_special_string_placeholders)

            pattern_insts_without_nonterminal_string_placeholders = self._apply_filters(
                pattern_insts_without_nonterminal_string_placeholders, filters, 1, tries)

            logger.debug("Found %d instantiations of pattern after instantiating nonterminal string placeholders",
                         len(pattern_insts_without_nonterminal_string_placeholders))

            if (pattern_insts_without_special_string_placeholders !=
                    pattern_insts_without_nonterminal_string_placeholders):
                pattern_insts_without_nonterminal_string_placeholders = self._filter_partial_instantiations(
                    pattern_insts_without_nonterminal_string_placeholders, tries)
                logger.debug("%d instantiations remain after filtering",
                             len(pattern_insts_without_nonterminal_string_placeholders))

            # 5. String placeholders
            if not any(isinstance(ph, StringPlaceholderVariableTypes) for ph in get_placeholders(pattern)):
                pattern_insts_without_string_placeholders = pattern_insts_without_nonterminal_string_placeholders
            else:
                pattern_insts_without_string_placeholders = self._instantiate_string_placeholders(
                    pattern_insts_without_nonterminal_string_placeholders, tries)

                logger.debug("Found %d instantiations of pattern after instantiating string placeholders",
                             len(pattern_insts_without_string_placeholders))

            assert all(not get_placeholders(candidate)
                       for candidate in pattern_insts_without_string_placeholders)

            result.update(pattern_insts_without_string_placeholders)

        return result

    def _filter_partial_instantiations(
            self,
            formulas: Iterable[language.Formula],
            tries: Iterable[datrie.Trie]) -> Set[language.Formula]:
        result: Set[language.Formula] = set()

        for pattern in formulas:
            for trie in tries:
                if not approximately_evaluate_abst_for(
                        pattern,
                        self.grammar,
                        self.graph,
                        {language.Constant("start", "<start>"): trie[path_to_trie_key(())]},
                        trie).is_false():
                    result.add(pattern)
                    break

        return result

    def _apply_filters(
            self,
            formulas: Set[language.Formula],
            filters: List['PatternInstantiationFilter'],
            order: int,
            tries: List[datrie.Trie],
            parallel: bool = False) -> Set[language.Formula]:
        # TODO: Check when parallel evaluation makes sense. In some cases, like
        #       "test_learn_from_islearn_patterns_file," the overhead of parallel
        #       evaluation seems to bee too high, in other cases (which?), it seems to work.
        for pattern_filter in filters:
            if not pattern_filter.order() == order:
                continue

            if parallel:
                formulas = list(formulas)

                with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
                    eval_results = list(pool.map(
                        lambda pattern_inst: int(pattern_filter.predicate(pattern_inst, tries)),
                        formulas,
                    ))

                formulas = set(itertools.compress(formulas, eval_results))
            else:
                formulas = {
                    pattern_inst for pattern_inst in formulas
                    if pattern_filter.predicate(pattern_inst, tries)
                }

            logger.debug("%d instantiations remaining after filter '%s'",
                         len(formulas),
                         pattern_filter.name)

        return formulas

    def _generate_more_inputs(self):
        logger.info(
            "Starting with %d positive, and %d negative samples.",
            len(self.positive_examples),
            len(self.negative_examples)
        )

        if not self.positive_examples:
            ne_before = len(self.negative_examples)
            pe_before = len(self.positive_examples)
            self._generate_sample_inputs()

            logger.info(
                "Generated %d additional positive, and %d additional negative samples (by grammar fuzzing).",
                len(self.positive_examples) - pe_before,
                len(self.negative_examples) - ne_before
            )

        assert len(self.positive_examples) > 0, "Cannot learn without any positive examples!"
        pe_before = len(self.positive_examples)
        ne_before = len(self.negative_examples)
        mutation_fuzzer = MutationFuzzer(self.grammar, self.positive_examples, self.prop, k=self.k)
        for inp in mutation_fuzzer.run(
                num_iterations=min(
                    30,
                    max(self.target_number_positive_samples - pe_before, 30),
                    max(self.target_number_negative_samples - ne_before, 30)
                ),
                alpha=.1, yield_negative=True):
            if self.prop(inp) and not tree_in(inp, self.positive_examples):
                self.positive_examples.append(inp)
            elif not self.prop(inp) and not tree_in(inp, self.negative_examples):
                self.negative_examples.append(inp)

        logger.info(
            "Generated %d additional positive, and %d additional negative samples (by mutation fuzzing).",
            len(self.positive_examples) - pe_before,
            len(self.negative_examples) - ne_before
        )

        if (len(self.positive_examples) < self.target_number_positive_samples or
                len(self.negative_examples) < self.target_number_negative_samples):
            ne_before = len(self.negative_examples)
            pe_before = len(self.positive_examples)
            self._generate_sample_inputs()

            # logger.debug(
            #     "Positive examples:\n%s",
            #     "\n".join(map(str, self.positive_examples)))

            logger.info(
                "Generated %d additional positive, and %d additional negative samples (by grammar fuzzing).",
                len(self.positive_examples) - pe_before,
                len(self.negative_examples) - ne_before
            )

        self.positive_examples = self._sort_inputs(
            self.positive_examples,
            self.filter_inputs_for_learning_by_kpaths,
            more_paths_weight=1.5,
            smaller_inputs_weight=1.0,
        )[:self.target_number_positive_samples]

        self.negative_examples = \
            self._sort_inputs(
                self.negative_examples,
                self.filter_inputs_for_learning_by_kpaths,
                more_paths_weight=2.0,
                smaller_inputs_weight=1.0
            )[:self.target_number_negative_samples]

        logger.info(
            "Reduced positive / negative samples to subsets of %d / %d samples based on k-path coverage.",
            len(self.positive_examples),
            len(self.negative_examples),
        )

        logger.debug(
            "Positive examples:\n%s",
            "\n".join(map(str, self.positive_examples)))

    def _generate_sample_inputs(
            self,
            num_tries: int = 100) -> None:
        fuzzer = isla.fuzzer.GrammarCoverageFuzzer(self.grammar)
        if (len(self.positive_examples) < self.target_number_positive_samples or
                len(self.negative_examples) < self.target_number_negative_samples):
            i = 0
            while ((len(self.positive_examples) < self.target_number_positive_samples
                    or len(self.negative_examples) < self.target_number_negative_samples)
                   and i < num_tries):
                i += 1

                inp = fuzzer.expand_tree(language.DerivationTree("<start>", None))

                if self.prop(inp) and not tree_in(inp, self.positive_examples):
                    self.positive_examples.append(inp)
                elif not self.prop(inp) and not tree_in(inp, self.negative_examples):
                    self.negative_examples.append(inp)

    def _generate_counter_examples_from_formulas(
            self,
            formulas: Iterable[language.Formula],
            desired_number_counter_examples: int = 50,
            num_tries: int = 100) -> Set[language.DerivationTree]:
        result: Set[language.DerivationTree] = set()

        solvers = {
            formula: ISLaSolver(self.grammar, formula, enforce_unique_trees_in_queue=False).solve()
            for formula in formulas}

        i = 0
        while len(result) < desired_number_counter_examples and i < num_tries:
            # with pmp.ProcessingPool(processes=pmp.cpu_count()) as pool:
            #     result.update({
            #         inp
            #         for inp in e_assert_present(pool.map(lambda f: next(solvers[f]), formulas))
            #         if not prop(inp)
            #     })
            for formula in formulas:
                for _ in range(desired_number_counter_examples):
                    inp = next(solvers[formula])
                    if not self.prop(inp):
                        result.add(inp)

            i += 1

        return result

    def _sort_inputs(
            self,
            inputs: Iterable[language.DerivationTree],
            filter_inputs_for_learning_by_kpaths: bool,
            more_paths_weight: float = 1.0,
            smaller_inputs_weight: float = 0.0) -> List[language.DerivationTree]:
        assert more_paths_weight or smaller_inputs_weight
        inputs = set(inputs)
        result: List[language.DerivationTree] = []

        tree_paths = {
            inp: {
                path for path in self.graph.k_paths_in_tree(inp.to_parse_tree(), self.k)
                if (not isinstance(path[-1], gg.TerminalNode) or
                    (not isinstance(path[-1], gg.TerminalNode) and len(path[-1].symbol) > 1))
            } for inp in inputs}

        covered_paths: Set[Tuple[gg.Node, ...]] = set([])
        max_len_input = max(len(inp) for inp in inputs)

        def uncovered_paths(inp: language.DerivationTree) -> Set[Tuple[gg.Node, ...]]:
            return {path for path in tree_paths[inp] if path not in covered_paths}

        def sort_by_paths_key(inp: language.DerivationTree) -> float:
            return len(uncovered_paths(inp))

        def sort_by_length_key(inp: language.DerivationTree) -> float:
            return len(inp)

        def sort_by_paths_and_length_key(inp: language.DerivationTree) -> float:
            return weighted_geometric_mean(
                [len(uncovered_paths(inp)), max_len_input - len(inp)],
                [more_paths_weight, smaller_inputs_weight])

        if not more_paths_weight:
            key = sort_by_length_key
        elif not smaller_inputs_weight:
            key = sort_by_paths_key
        else:
            key = sort_by_paths_and_length_key

        while inputs:
            inp = sorted(inputs, key=key, reverse=True)[0]
            inputs.remove(inp)
            uncovered = uncovered_paths(inp)

            if filter_inputs_for_learning_by_kpaths and not uncovered:
                continue

            covered_paths.update(uncovered)
            result.append(inp)

        return result

    def _instantiate_nonterminal_placeholders(
            self,
            pattern: language.Formula,
            input_reachability_relation: Set[Tuple[str, str]],
            _instantiations: Optional[List[Dict[NonterminalPlaceholderVariable, language.BoundVariable]]] = None
    ) -> Set[language.Formula]:
        if _instantiations:
            instantiations = _instantiations
        else:
            instantiations = self._instantiations_for_placeholder_variables(pattern, input_reachability_relation)

        result: Set[language.Formula] = {
            pattern.substitute_variables(instantiation)
            for instantiation in instantiations
        }

        return result

    def _instantiations_for_placeholder_variables(
            self,
            pattern: language.Formula,
            input_reachability_relation: Set[Tuple[str, str]]
    ) -> List[Dict[NonterminalPlaceholderVariable, language.BoundVariable]]:
        instantiations: List[Dict[NonterminalPlaceholderVariable, language.BoundVariable]] = []

        in_visitor = InVisitor()
        pattern.accept(in_visitor)
        variable_chains: Set[Tuple[language.Variable, ...]] = connected_chains(in_visitor.result)
        start_const = extract_top_level_constant(pattern)
        assert all(chain[-1] == start_const for chain in variable_chains)
        reachable = {fr: [to for fr_, to in input_reachability_relation
                          if fr == fr_ and to not in self.exclude_nonterminals] for fr in self.grammar}

        # Combine variable chains to a tree structure to avoid conflicting instantiations.
        InstantiationTree = Tree[Tuple[language.Variable, Optional[str]]]
        initial_tree: InstantiationTree = tree_from_paths(
            [list(reversed(chain)) for chain in variable_chains])

        # We basically perform a BFS over the partially instantiated trees and
        # instantiate children based on the parent values.
        stack: List[Tuple[
            List[InstantiationTree],
            Dict[NonterminalPlaceholderVariable, language.BoundVariable]]] = \
            [([tree for _, tree in tree_paths(initial_tree)], {})]
        while stack:
            remaining_subtrees, inst_map = stack.pop()
            if not remaining_subtrees:
                instantiations.append(inst_map)
                continue

            parent_variable: PlaceholderVariable
            children: List[InstantiationTree]
            (parent_variable, children), *remaining_subtrees = remaining_subtrees

            if not children:
                stack.append((remaining_subtrees, inst_map))
                continue

            if isinstance(parent_variable, PlaceholderVariable):
                assert parent_variable in inst_map
                parent_instantiation = inst_map[parent_variable].n_type
            else:
                parent_instantiation = parent_variable.n_type

            reachable_nonterminals = reachable[parent_instantiation]
            if not reachable_nonterminals:
                continue

            for instantiation in itertools.product(*[reachable_nonterminals for _ in range(len(children))]):
                assert all(child[0] not in inst_map for child in children)
                child_insts = {
                    child[0]: language.BoundVariable(child[0].name, instantiation[idx])
                    for idx, child in enumerate(children)}
                stack.append((remaining_subtrees, inst_map | child_insts))

        return instantiations

        # for variable_chain in variable_chains:
        #     nonterminal_sequences: Set[Tuple[str, ...]] = set([])
        #
        #     partial_sequences: Set[Tuple[str, ...]] = {("<start>",)}
        #     while partial_sequences:
        #         partial_sequence = next(iter(partial_sequences))
        #         partial_sequences.remove(partial_sequence)
        #         if len(partial_sequence) == len(variable_chain):
        #             nonterminal_sequences.add(tuple(reversed(partial_sequence)))
        #             continue
        #
        #         partial_sequences.update({
        #             partial_sequence + (to_nonterminal,)
        #             for to_nonterminal in reachable[partial_sequence[-1]]
        #             if to_nonterminal not in self.exclude_nonterminals})
        #
        #     new_instantiations: List[Dict[NonterminalPlaceholderVariable, language.BoundVariable]] = [
        #         {variable: language.BoundVariable(variable.name, nonterminal_sequence[idx])
        #          for idx, variable in enumerate(variable_chain[:-1])}
        #         for nonterminal_sequence in nonterminal_sequences]
        #
        #     if not instantiations:
        #         instantiations = new_instantiations
        #     else:
        #         previous_instantiations = instantiations
        #         instantiations = []
        #         for previous_instantiation in previous_instantiations:
        #             for new_instantiation in new_instantiations:
        #                 # NOTE: There might be clashes, since multiple chains are generated if there is
        #                 #       more than one variable in a match expression. We have to first check if
        #                 #       two instantiations happen to conform to each other.
        #                 if any(new_instantiation[key].n_type != previous_instantiation[key].n_type for key in
        #                        set(new_instantiation.keys()).intersection(set(previous_instantiation.keys()))):
        #                     continue
        #
        #                 instantiations.append(previous_instantiation | new_instantiation)
        #
        # return instantiations

    def _instantiate_mexpr_placeholders(
            self,
            inst_patterns: Set[language.Formula]) -> Set[language.Formula]:
        def quantified_formulas_with_mexpr_phs(formula: language.Formula) -> Set[language.QuantifiedFormula]:
            return cast(Set[language.QuantifiedFormula], set(language.FilterVisitor(
                lambda f: (isinstance(f, language.QuantifiedFormula) and
                           f.bind_expression is not None and
                           isinstance(f.bind_expression, AbstractBindExpression) and
                           isinstance(f.bind_expression.bound_elements[0], MexprPlaceholderVariable))).collect(
                formula)))

        if not any(quantified_formulas_with_mexpr_phs(pattern) for pattern in inst_patterns):
            return inst_patterns

        result: Set[language.Formula] = set([])
        stack: List[language.Formula] = list(inst_patterns)
        while stack:
            pattern = stack.pop()
            qfd_formulas_with_mexpr_phs = quantified_formulas_with_mexpr_phs(pattern)
            if not qfd_formulas_with_mexpr_phs:
                result.add(pattern)
                continue

            for qfd_formula_w_mexpr_phs in qfd_formulas_with_mexpr_phs:
                mexpr_ph: MexprPlaceholderVariable = cast(
                    AbstractBindExpression,
                    qfd_formula_w_mexpr_phs.bind_expression).bound_elements[0]
                in_nonterminal = qfd_formula_w_mexpr_phs.bound_variable.n_type
                nonterminal_types: Tuple[str, ...] = tuple([var.n_type for var in mexpr_ph.variables])

                for mexpr_strings in self._infer_mexpr(in_nonterminal, nonterminal_types):
                    def replace_with_var(elem: str) -> str | language.Variable:
                        try:
                            return next(var for idx, var in enumerate(mexpr_ph.variables)
                                        if elem == var.n_type.replace(">", f"-{hash((var.n_type, idx))}>"))
                        except StopIteration:
                            return elem

                    mexpr_elements = [replace_with_var(element) for element in mexpr_strings]

                    constructor = (
                        language.ForallFormula if isinstance(qfd_formula_w_mexpr_phs, language.ForallFormula)
                        else language.ExistsFormula)

                    new_formula = constructor(
                        qfd_formula_w_mexpr_phs.bound_variable,
                        qfd_formula_w_mexpr_phs.in_variable,
                        qfd_formula_w_mexpr_phs.inner_formula,
                        language.BindExpression(*mexpr_elements))

                    stack.append(language.replace_formula(pattern, qfd_formula_w_mexpr_phs, new_formula))

        return result

    @lru_cache(maxsize=None)
    def _infer_mexpr(
            self,
            in_nonterminal: str,
            nonterminal_types: Tuple[str, ...]) -> Set[Tuple[str, ...]]:
        assert all(self.graph.reachable(in_nonterminal, target) for target in nonterminal_types)

        result: Set[Tuple[str, ...]] = set()

        candidate_trees: List[ParseTree] = [(in_nonterminal, None)]
        for i in range(self.mexpr_expansion_limit + 1):
            if not candidate_trees:
                break

            old_candidates = list(candidate_trees)
            candidate_trees = []
            for tree in old_candidates:
                # If the candidate tree has the shape
                #
                #          qfd_nonterminal
                #                 |
                #         other_nonterminal
                #                 |
                #          ... subtree ...
                #
                # then we could as well quantify over `other_nonterminal`. Consequently,
                # we prune this candidate. The other option is very likely also among
                # inst_patterns.

                if (tree[1] is not None and
                        len(tree[1]) == 1 and
                        tree[1][0][1]):
                    continue

                nonterminal_occurrences: List[Tuple[Path, ...]] = []
                trie = trie_from_parse_tree(tree)
                stack: List[Tuple[Tuple[Path, ...], Tuple[str], str]] = \
                    [((), nonterminal_types, path_to_trie_key(()))]
                while stack:
                    matches, remaining_types, trie_key = stack.pop(0)
                    if not remaining_types:
                        nonterminal_occurrences.append(matches)

                    cur_path, cur_tree = trie[trie_key]

                    if cur_tree[0] == remaining_types[0]:
                        remaining_types = remaining_types[1:]
                        matches = matches + (cur_path,)

                        if not remaining_types:
                            nonterminal_occurrences.append(matches)
                            continue

                        # For the next trie key, skip the whole subtree of cur_path
                        if not cur_tree[1]:
                            next_key = next_trie_key(trie, trie_key)
                        else:
                            suffixes = list(filter(None, trie.suffixes(trie_key)))
                            assert suffixes
                            next_key = next_trie_key(trie, trie_key + suffixes[-1])
                    else:
                        next_key = next_trie_key(trie, trie_key)

                    if next_key is not None:
                        stack.append((matches, remaining_types, next_key))

                for matching_seq in nonterminal_occurrences:
                    assert len(matching_seq) == len(nonterminal_types)

                    # We change node labels to be able to correctly identify variable positions in the tree.
                    bind_expr_tree = tree
                    for idx, path in enumerate(matching_seq):
                        ntype = nonterminal_types[idx]
                        bind_expr_tree = replace_path(
                            bind_expr_tree,
                            path,
                            (ntype.replace(">", f"-{hash((ntype, idx))}>"), None))

                    # We prune too specific leaves of the tree: Each node that does not contain
                    # a variable node as a child, and is an immediate child of a node containing
                    # a variable node, is pruned away.
                    non_var_paths = [path for path, _ in tree_leaves(bind_expr_tree)
                                     if path not in matching_seq]

                    for path in non_var_paths:
                        if len(path) < 2:
                            continue

                        while len(path) > 1 and not any(
                                len(path[:-1]) < len(var_path) and
                                var_path[:len(path[:-1])] == path[:-1]
                                for var_path in matching_seq):
                            path = path[:-1]

                        if len(path) > 0 and not any(
                                len(path) < len(var_path) and
                                var_path[:len(path)] == path
                                for var_path in matching_seq):
                            bind_expr_tree = replace_path(
                                bind_expr_tree,
                                path,
                                (get_subtree(bind_expr_tree, path)[0], None))

                    result.add(tuple([tree[0] for _, tree in tree_leaves(bind_expr_tree)]))

                expanded_trees = expand_tree(tree, self.canonical_grammar)
                if self.max_nonterminals_in_mexpr is not None:
                    expanded_trees = [
                        t for t in expanded_trees
                        if (len([True for _, leaf in tree_leaves(t) if is_nonterminal(leaf[0])]) <=
                            self.max_nonterminals_in_mexpr)
                    ]
                candidate_trees.extend(expanded_trees)

        return result

    def _instantiate_nonterminal_string_placeholders(
            self,
            inst_patterns: Set[language.Formula]) -> Set[language.Formula]:
        if all(not isinstance(placeholder, NonterminalStringPlaceholderVariable)
               for inst_pattern in inst_patterns
               for placeholder in get_placeholders(inst_pattern)):
            return inst_patterns

        result: Set[language.Formula] = set([])
        for inst_pattern in inst_patterns:
            for nonterminal in self.grammar:
                def replace_placeholder_by_nonterminal_string(
                        subformula: language.Formula) -> language.Formula | bool:
                    if (not isinstance(subformula, language.SemanticPredicateFormula) and
                            not isinstance(subformula, language.StructuralPredicateFormula)):
                        return False

                    if not any(isinstance(arg, NonterminalStringPlaceholderVariable) for arg in subformula.args):
                        return False

                    new_args = [
                        nonterminal if isinstance(arg, NonterminalStringPlaceholderVariable)
                        else arg
                        for arg in subformula.args]

                    constructor = (
                        language.StructuralPredicateFormula
                        if isinstance(inst_pattern, language.StructuralPredicateFormula)
                        else language.SemanticPredicateFormula)

                    return constructor(subformula.predicate, *new_args)

                result.add(language.replace_formula(inst_pattern, replace_placeholder_by_nonterminal_string))

        return result

    def __instantiate_special_predicate_string_placeholders(
            self,
            inst_patterns: Set[language.Formula],
            tries: List[datrie.Trie]) -> Set[language.Formula]:
        result: Set[language.Formula] = set([])

        for pattern in inst_patterns:
            nth_predicates = language.FilterVisitor(
                lambda f: (isinstance(f, language.StructuralPredicateFormula) and
                           f.predicate == isla_predicates.NTH_PREDICATE and
                           isinstance(f.args[0], StringPlaceholderVariable))
            ).collect(pattern)

            if not nth_predicates:
                result.add(pattern)

            # We expect that we don't count in "trivial" nonterminals, i.e., those that immediately
            # unfold to individual characters.
            if any(self._reachable_characters(nth_predicate.args[1].n_type) for nth_predicate in nth_predicates):
                continue

            # For each `nth(<STRING>, elem, container)` add as instantiations all indices
            # of occurrences of `elem`'s nonterminal withint `container`
            sub_result: List[language.Formula] = [pattern]
            nth_predicate: language.StructuralPredicateFormula
            for nth_predicate in nth_predicates:
                elem_nonterminal = nth_predicate.args[1].n_type
                container_nonterminal = nth_predicate.args[2].n_type
                indices: Set[int] = set([])
                for trie in tries:
                    container_tries = [
                        get_subtrie(trie, path_key) for path_key, (path, subtree) in trie.items()
                        if subtree.value == container_nonterminal]

                    for container_trie in container_tries:
                        num_occs = len([
                            subtree for _, subtree in container_trie.values()
                            if subtree.value == elem_nonterminal])

                        indices.update(list(range(1, num_occs + 1)))

                for partial_result in list(sub_result):
                    sub_result.remove(partial_result)

                    def replace_placeholder_by_idx(
                            index: int, subformula: language.Formula) -> language.Formula | bool:
                        if (not isinstance(subformula, language.StructuralPredicateFormula) or
                                subformula.predicate != isla_predicates.NTH_PREDICATE or
                                not isinstance(subformula.args[0], StringPlaceholderVariable)):
                            return False

                        return language.StructuralPredicateFormula(
                            subformula.predicate, str(index), subformula.args[1], subformula.args[2]
                        )

                    sub_result.extend([
                        language.replace_formula(pattern, functools.partial(replace_placeholder_by_idx, index))
                        for index in indices])

            result.update(sub_result)

        return result

    @lru_cache(maxsize=100)
    def _reachable_characters(self, symbol: str) -> Optional[Set[str]]:
        if not is_nonterminal(symbol):
            return {symbol}

        if any(len(expansion) > 1 for expansion in self.canonical_grammar[symbol]):
            return None

        expansion_elements = {
            element for expansion in self.canonical_grammar[symbol]
            for element in expansion
        }

        if all(len(element) == 1 and not is_nonterminal(element)
               for element in expansion_elements):
            return expansion_elements

        children_results = [
            self._reachable_characters(element)
            for element in expansion_elements
        ]

        if any(child_result is None for child_result in children_results):
            return None

        return set(functools.reduce(set.__or__, children_results))

    def _instantiate_string_placeholders(
            self,
            inst_patterns: Set[language.Formula],
            tries: List[datrie.Trie]):
        result: Set[language.Formula] = set([])
        string_placeholder_insts = self._get_string_placeholder_instantiations(inst_patterns, tries)

        for formula in string_placeholder_insts:
            if not string_placeholder_insts[formula]:
                result.add(formula)
                continue

            # We make instantiations for DSTRINGS placeholders map to *sets of tuples* of strings instead of
            # sets of strings. To prevent explosion, we apply a simple heuristic: We consider numbers and
            # single letters as irrelevant. This holds for our use cases (e.g., considering 'xmlns', 'sqrt',
            # and '*') but can be problematic in cases where something like 'f' should be considered. However,
            # some reduction *has* to be done, and this works for our use cases and seems to be sensile in
            # general (protected/pre-defined identifiers are rarely numbers or single characters).
            insts = {
                ph:
                    s if isinstance(ph, StringPlaceholderVariable)
                    else (s_ := {elem for elem in s if not is_int(elem) and elem not in string.ascii_letters},
                          {c for k in range(1, len(s_) + 1) for c in tuple(itertools.combinations(s_, k))})[-1]
                for ph, s in string_placeholder_insts[formula].items()}

            string_ph_insts: List[Dict[StringPlaceholderVariable, str | Set[str]]] = \
                dict_of_lists_to_list_of_dicts(insts)

            for instantiation in string_ph_insts:
                # TODO: We have to account for sets in instantiations in the abstract
                #       evaluation, or expand the formula before. Probably try the former.
                if any(
                        not approximately_evaluate_abst_for(
                            formula,
                            self.grammar,
                            self.graph,
                            {language.Constant("start", "<start>"): trie[path_to_trie_key(())]} | instantiation,
                            trie
                        ).is_false()
                        for trie in tries):
                    instantiated_formula = formula
                    for ph, inst in instantiation.items():
                        instantiated_formula = language.replace_formula(
                            instantiated_formula,
                            functools.partial(substitute_string_placeholder, ph, inst))
                    result.add(instantiated_formula)

        return result

    def _get_string_placeholder_instantiations(
            self,
            inst_patterns: Set[language.Formula],
            tries: List[datrie.Trie]) -> Dict[language.Formula, Dict[StringPlaceholderVariable, Set[str]]]:
        if all(not isinstance(placeholder, StringPlaceholderVariableTypes)
               for inst_pattern in inst_patterns
               for placeholder in get_placeholders(inst_pattern)):
            return dict.fromkeys(inst_patterns)

        # TODO: Record fragments inside their *contexts*, and only instantiate in the right context.
        #       An <ID> in at the place of a function name is different than an <ID> at the place of
        #       a variable.
        fragments: Dict[str, Set[str]] = {nonterminal: set([]) for nonterminal in self.grammar}
        for trie in tries:
            remaining_paths: List[Tuple[Path, language.DerivationTree]] = list(
                sorted(list(trie.values()), key=lambda p: len(p[0])))

            handled_paths: Dict[str, Set[Path]] = {nonterminal: set([]) for nonterminal in self.grammar}
            while remaining_paths:
                path, tree = remaining_paths.pop(0)
                if not is_nonterminal(tree.value):
                    continue

                tree_string = str(tree)
                if not tree_string:
                    continue

                # NOTE: We exclude substrings from fragments; e.g., if we have a <digits>
                #       "1234", don't include the <digits> "34". This might lead
                #       to imprecision, but otherwise the search space tends to explode.
                if any(len(opath) < len(path) and opath == path[:len(opath)]
                       for opath in handled_paths[tree.value]):
                    continue

                handled_paths[tree.value].add(path)
                fragments[tree.value].add(tree_string)
                fragments[tree.value].add(str(len(tree_string)))

                # For strings representing floats, we also include the rounded Integers.
                if is_float(tree_string) and not is_int(tree_string):
                    fragments[tree.value].add(str(int(float(tree_string))))
                    fragments[tree.value].add(str(int(float(tree_string)) + 1))

        logger.debug(
            "Extracted %d language fragments from sample inputs",
            sum(len(value) for value in fragments.values()))

        result: Dict[language.Formula, Dict[StringPlaceholderVariableTypes, Set[str]]] = {}
        for inst_pattern in inst_patterns:
            def extract_instantiations(subformula: language.Formula) -> None:
                nonlocal result

                if not any(isinstance(arg, StringPlaceholderVariableTypes) for arg in subformula.free_variables()):
                    return

                non_ph_vars = {v for v in subformula.free_variables() if not isinstance(v, PlaceholderVariable)}

                insts: Set[str] = set(functools.reduce(
                    set.__or__, [fragments.get(var.n_type, set([])) for var in non_ph_vars]))

                ph_vars = {v for v in subformula.free_variables() if isinstance(v, StringPlaceholderVariableTypes)}
                for ph in ph_vars:
                    result.setdefault(inst_pattern, {}).setdefault(ph, set([])).update(insts)

            class InstantiationVisitor(language.FormulaVisitor):
                def visit_smt_formula(self, formula: language.SMTFormula):
                    extract_instantiations(formula)

                def visit_predicate_formula(self, formula: language.StructuralPredicateFormula):
                    extract_instantiations(formula)

                def visit_semantic_predicate_formula(self, formula: language.SemanticPredicateFormula):
                    extract_instantiations(formula)

            inst_pattern.accept(InstantiationVisitor())

        return result


def substitute_string_placeholder(
        ph: PlaceholderVariable,
        inst: str | tuple[str, ...],
        formula: language.Formula) -> language.Formula | bool:
    # A tuple of instantiations indicates that a disjunction of formulas shall be
    # returned, including one instantiated formula for each instantiation.

    if not any(v == ph for v in formula.free_variables()
               if isinstance(v, StringPlaceholderVariableTypes)):
        return False

    def perform_single_inst(inst_str: str) -> Optional[language.Formula]:
        if isinstance(formula, language.SMTFormula):
            return language.SMTFormula(cast(z3.BoolRef, z3_subst(formula.formula, {
                ph.to_smt(): z3.StringVal(inst_str)
            })), *[v for v in formula.free_variables() if v != ph])
        elif isinstance(formula, language.StructuralPredicateFormula):
            return language.StructuralPredicateFormula(
                formula.predicate,
                *[inst_str if arg == ph else arg for arg in formula.args]
            )
        elif isinstance(formula, language.SemanticPredicateFormula):
            return language.SemanticPredicateFormula(
                formula.predicate,
                *[inst_str if arg == ph else arg for arg in formula.args])

        return None

    if isinstance(inst, str):
        inst = inst,

    single_insts = list(map(perform_single_inst, inst))
    if any(single_inst is None for single_inst in single_insts):
        return False

    return functools.reduce(language.Formula.__or__, single_insts)


def approximately_evaluate_abst_for(
        formula: language.Formula,
        grammar: Grammar,
        graph: gg.GrammarGraph,
        assignments: Dict[language.Variable, Tuple[Path, language.DerivationTree] | str | Set[str]],
        trie: Optional[datrie.Trie] = None) -> ThreeValuedTruth:
    # TODO: Handle String placeholder variables in predicate formulas
    if isinstance(formula, language.SMTFormula):
        if any(isinstance(arg, PlaceholderVariable) and arg not in assignments
               for arg in formula.free_variables()):
            return ThreeValuedTruth.unknown()

        if any(var not in assignments for var in formula.free_variables()):
            # This happens for quantified formulas over integers: We simply
            # strip of the quantifier, and a variable remains.
            return ThreeValuedTruth.unknown()

        try:
            translation = evaluate_z3_expression(formula.formula)
            var_map: Dict[str, language.Variable] = {
                var.name: var
                for var in formula.free_variables()
            }

            # Expand multiple instantiations for `DisjunctiveStringsPlaceholderVariable`s
            split_assignments: List[Dict[language.Variable, Tuple[Path, language.DerivationTree] | str]] = [
                dict(t) for t in itertools.product(*[
                    ((var, val),) if not isinstance(var, DisjunctiveStringsPlaceholderVariable)
                    else tuple(itertools.product((var,), e_assert(val, lambda v: isinstance(v, tuple))))
                    for var, val in assignments.items()
                    if var in formula.free_variables()])]

            # The multiple assignments in the list are treated as a disjunction:
            # As soon as any assignment yields a `True` result, we return `True`.

            for single_assignment in split_assignments:
                args_instantiation = ()
                for arg in translation[0]:
                    var = var_map[arg]
                    if isinstance(var, StringPlaceholderVariableTypes):
                        args_instantiation += (single_assignment[var],)
                    else:
                        args_instantiation += (str(single_assignment[var][1]),)

                if translation[1](args_instantiation) if args_instantiation else translation[1]:
                    return ThreeValuedTruth.true()

            return ThreeValuedTruth.false()
        except DomainError:
            return ThreeValuedTruth.false()
        except NotImplementedError:
            return is_valid(z3.substitute(
                formula.formula,
                *tuple({z3.String(symbol.name): z3.StringVal(str(symbol_assignment[1]))
                        for symbol, symbol_assignment
                        in assignments.items()}.items())))

    elif isinstance(formula, language.NumericQuantifiedFormula):
        return approximately_evaluate_abst_for(formula.inner_formula, grammar, graph, assignments, trie)
    elif isinstance(formula, language.QuantifiedFormula):
        assert isinstance(formula.in_variable, language.Variable)
        assert formula.in_variable in assignments
        in_path, in_inst = assignments[formula.in_variable]

        if formula.bind_expression is None:
            sub_trie = get_subtrie(trie, in_path)

            new_assignments: List[Dict[language.Variable, Tuple[Path, language.DerivationTree]]] = []
            for path, subtree in sub_trie.values():
                if subtree.value == formula.bound_variable.n_type:
                    new_assignments.append({formula.bound_variable: (in_path + path, subtree)})
        elif isinstance(formula.bind_expression.bound_elements[0], MexprPlaceholderVariable):
            mexpr_placeholder = cast(MexprPlaceholderVariable, formula.bind_expression.bound_elements[0])

            # First, get all matches for the bound variables
            new_assignments: List[Dict[language.Variable, Tuple[Path, language.DerivationTree]]] = []
            for path, subtree in get_subtrie(trie, in_path).values():
                if subtree.value == formula.bound_variable.n_type:
                    new_assignments.append({formula.bound_variable: (in_path + path, subtree)})

            # Next, find subtrees below those matches matching the mexpr placeholder in the correct order
            for idx, placeholder_variable in enumerate(mexpr_placeholder.variables):
                if not idx:
                    for assignment in list(new_assignments):
                        new_assignments.remove(assignment)

                        sub_trie = get_subtrie(trie, assignment[formula.bound_variable][0])
                        matches = [
                            (in_path + path, subtree)
                            for path, subtree in sub_trie.values()
                            if subtree.value == placeholder_variable.n_type]
                        if not matches:
                            continue

                        new_assignments.extend([
                            assignment | {placeholder_variable: (path, subtree)}
                            for path, subtree in matches
                        ])
                else:
                    last_variable: NonterminalPlaceholderVariable = mexpr_placeholder.variables[idx - 1]
                    for assignment in list(new_assignments):
                        new_assignments.remove(assignment)

                        sub_trie = get_subtrie(trie, assignment[formula.bound_variable][0])
                        last_path: Path = assignment[last_variable][0]
                        for sub_path_key in sub_trie.keys(path_to_trie_key(last_path)):
                            del sub_trie[sub_path_key]

                        next_key = next_trie_key(sub_trie, path_to_trie_key(last_path))
                        while next_key is not None:
                            sub_sub_trie = get_subtrie(sub_trie, next_key)

                            matches = [
                                (last_path + path, subtree)
                                for path, subtree in sub_sub_trie.values()
                                if subtree.value == placeholder_variable.n_type]
                            if not matches:
                                continue

                            new_assignments.extend([
                                assignment | {placeholder_variable: (path, subtree)}
                                for path, subtree in matches
                            ])

                            for sub_path_key in sub_trie.keys(next_key):
                                del sub_trie[sub_path_key]

                            next_key = next_trie_key(sub_trie, path_to_trie_key(last_path))

            # For universal formulas, we only consider the first 3 new assignments
            # to save time. After match expression placeholders are instantiated,
            # we have the chance for a more precise check.
            assert all(formula.bound_variable in assignment for assignment in new_assignments)
            assert all(var in assignment for assignment in new_assignments for var in mexpr_placeholder.variables)
            if isinstance(formula, language.ForallFormula):
                new_assignments = new_assignments[:3]
        else:
            new_assignments = [
                {var: (in_path + path, tree) for var, (path, tree) in new_assignment.items()}
                for new_assignment in matches_for_quantified_formula(
                    formula, grammar, in_inst, {})]

        if not new_assignments:
            return ThreeValuedTruth.false()

        new_assignments = [
            new_assignment | assignments
            for new_assignment in new_assignments]

        if isinstance(formula, language.ExistsFormula):
            return ThreeValuedTruth.from_bool(any(
                not approximately_evaluate_abst_for(
                    formula.inner_formula, grammar, graph, new_assignment, trie).is_false()
                for new_assignment in new_assignments))
        else:
            return ThreeValuedTruth.from_bool(all(
                not approximately_evaluate_abst_for(
                    formula.inner_formula, grammar, graph, new_assignment, trie).is_false()
                for new_assignment in new_assignments))
    elif isinstance(formula, language.StructuralPredicateFormula):
        if any(isinstance(arg, PlaceholderVariable) for arg in formula.args):
            return ThreeValuedTruth.unknown()

        arg_insts = [
            arg if isinstance(arg, str)
            else next(path for path, subtree in trie.values() if subtree.id == arg.id)
            if isinstance(arg, language.DerivationTree)
            else assignments[arg][0]
            for arg in formula.args]

        return ThreeValuedTruth.from_bool(formula.predicate.evaluate(trie.get(path_to_trie_key(()))[1], *arg_insts))
    elif isinstance(formula, language.SemanticPredicateFormula):
        if any(isinstance(arg, PlaceholderVariable) for arg in formula.args):
            return ThreeValuedTruth.unknown()

        arg_insts = [arg if isinstance(arg, language.DerivationTree) or arg not in assignments
                     else assignments[arg][1]
                     for arg in formula.args]
        eval_res = formula.predicate.evaluate(graph, *arg_insts)

        if eval_res.true():
            return ThreeValuedTruth.true()
        elif eval_res.false():
            return ThreeValuedTruth.false()

        if not eval_res.ready() or not all(isinstance(key, language.Constant) for key in eval_res.result):
            # Evaluation resulted in a tree update; that is, the formula is satisfiable, but only
            # after an update of its arguments. This result happens when evaluating formulas during
            # solution search after instantiating variables with concrete trees.
            return ThreeValuedTruth.unknown()

        assignments.update({const: (tuple(), assgn) for const, assgn in eval_res.result.items()})
        return ThreeValuedTruth.true()
    elif isinstance(formula, language.NegatedFormula):
        return ThreeValuedTruth.not_(
            approximately_evaluate_abst_for(formula.args[0], grammar, graph, assignments, trie))
    elif isinstance(formula, language.ConjunctiveFormula):
        # Relaxation: Unknown is OK, only False is excluded.
        return ThreeValuedTruth.from_bool(all(
            not approximately_evaluate_abst_for(sub_formula, grammar, graph, assignments, trie).is_false()
            for sub_formula in formula.args))
    elif isinstance(formula, language.DisjunctiveFormula):
        return ThreeValuedTruth.from_bool(any(
            not approximately_evaluate_abst_for(sub_formula, grammar, graph, assignments, trie).is_false()
            for sub_formula in formula.args))
    else:
        raise NotImplementedError()


class PatternInstantiationFilter(ABC):
    def __init__(self, name: str):
        self.name = name

    def order(self) -> int:
        """Currently: If 0, can be executed after all nonterminal types in quantifiers are fixed.
        If 1, can only be executed after also all nonterminal string placeholders (in predicates)
        have been instantiated. If 2, can only be executed after also all string placeholders
        have been instantiated."""
        raise NotImplementedError()

    def __eq__(self, other):
        return isinstance(other, PatternInstantiationFilter) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def predicate(self, formula: language.Formula, tries: List[datrie.Trie]) -> bool:
        raise NotImplementedError()


class NonterminalStringInCountPredicatesFilter(PatternInstantiationFilter):
    def __init__(self, graph: gg.GrammarGraph, input_reachability_relation: Set[Tuple[str, str]]):
        super().__init__("Nonterminal String in `count` Predicates Filter")
        self.graph = graph
        self.input_reachability_relation = input_reachability_relation

    def order(self) -> int:
        return 1

    def reachable(self, from_nonterminal: str, to_nonterminal: str) -> bool:
        return reachable(self.graph, from_nonterminal, to_nonterminal)

    def reachable_in_inputs(self, from_nonterminal: str, to_nonterminal: str) -> bool:
        return (from_nonterminal, to_nonterminal) in self.input_reachability_relation

    def predicate(self, formula: language.Formula, _: List[datrie.Trie]) -> bool:
        # In `count(elem, nonterminal, num)` occurrences
        # 1. the nonterminal must be reachable from the nonterminal type of elem. We
        #    consider reachability as defined by the sample inputs, not the grammar,
        #    to increase precision,
        # 2. it must be possible that the nonterminal occurs a variable number of times
        #    in the subgrammar of elem's nonterminal type.

        count_predicates: List[language.SemanticPredicateFormula] = cast(
            List[language.SemanticPredicateFormula],
            language.FilterVisitor(
                lambda f: isinstance(f, language.SemanticPredicateFormula) and
                          f.predicate == isla_predicates.COUNT_PREDICATE).collect(formula))

        if not count_predicates:
            return True

        def reachable_variable_number_of_times(start_nonterminal: str, needle_nonterminal: str) -> bool:
            return any(self.reachable_in_inputs(start_nonterminal, other_nonterminal) and
                       self.reachable(other_nonterminal, other_nonterminal) and
                       self.reachable_in_inputs(other_nonterminal, needle_nonterminal)
                       for other_nonterminal in self.graph.grammar)

        return any(
            reachable_variable_number_of_times(count_predicate.args[0].n_type, count_predicate.args[1])
            for count_predicate in count_predicates
        )


class InVisitor(language.FormulaVisitor):
    def __init__(self):
        self.result: Set[Tuple[language.Variable, language.Variable]] = set()

    def visit_exists_formula(self, formula: language.ExistsFormula):
        self.handle(formula)

    def visit_forall_formula(self, formula: language.ForallFormula):
        self.handle(formula)

    def handle(self, formula: language.QuantifiedFormula):
        self.result.add((formula.bound_variable, formula.in_variable))
        if formula.bind_expression:
            if any(isinstance(elem, MexprPlaceholderVariable)
                   for elem in formula.bind_expression.bound_elements):
                phs = [elem for elem in formula.bind_expression.bound_elements
                       if isinstance(elem, MexprPlaceholderVariable)]
                assert len(phs) == 1
                ph = cast(MexprPlaceholderVariable, phs[0])
                for nonterminal_placeholder in ph.variables:
                    self.result.add((nonterminal_placeholder, formula.bound_variable))
            else:
                self.result.update({
                    (var, formula.bound_variable)
                    for var in formula.bind_expression.bound_elements
                    if isinstance(var, language.BoundVariable)
                       and not isinstance(var, language.DummyVariable)
                })


def get_placeholders(formula: language.Formula) -> Set[PlaceholderVariable]:
    placeholders = {var for var in language.VariablesCollector.collect(formula)
                    if isinstance(var, PlaceholderVariable)}
    supported_placeholder_types = {
        NonterminalPlaceholderVariable,
        NonterminalStringPlaceholderVariable,
        StringPlaceholderVariable,
        DisjunctiveStringsPlaceholderVariable,
    }

    assert all(any(isinstance(ph, t) for t in supported_placeholder_types) for ph in placeholders), \
        "Only " + ", ".join(map(lambda t: t.__name__, supported_placeholder_types)) + " supported so far."

    return placeholders


def extract_top_level_constant(candidate):
    return next(
        (c for c in language.VariablesCollector.collect(candidate)
         if isinstance(c, language.Constant) and not c.is_numeric()))


class PatternRepository:
    def __init__(self, data: Dict[str, List[Dict[str, str]]]):
        self.groups: Dict[str, Dict[str, language.Formula]] = {}
        for group_name, elements in data.items():
            for entry in elements:
                name = entry["name"]
                constraint = parse_abstract_isla(entry["constraint"])
                self.groups.setdefault(group_name, {})[name] = constraint

    def __getitem__(self, item: str) -> Set[language.Formula]:
        if item in self.groups:
            return set(self.groups[item].values())

        for elements in self.groups.values():
            if item in elements:
                return {elements[item]}

        logger.warning(f"Could not find pattern for query {item}.")
        return set([])

    def __contains__(self, item: str) -> bool:
        return bool(self[item])

    def __len__(self):
        return sum([len(group.values()) for group in self.groups.values()])

    def get_all(self, but: Iterable[str] = tuple()) -> Set[language.Formula]:
        exclude: Set[language.Formula] = set([])
        for excluded in but:
            exclude.update(self[excluded])

        all_patterns: Set[language.Formula] = set(
            functools.reduce(set.__or__, [set(d.values()) for d in self.groups.values()]))

        return all_patterns - exclude

    def __str__(self):
        result = ""
        for group in self.groups:
            for name in self.groups[group]:
                result += f"[[{group}]]\n\n"
                result += f'name = "{name}"\n'
                result += "constraint = '''\n"
                result += AbstractISLaUnparser(self.groups[group][name]).unparse()
                result += "\n'''"
                result += "\n\n"

        return result.strip()


def create_input_reachability_relation(inputs: Iterable[language.DerivationTree]) -> Set[Tuple[str, str]]:
    nonterminal_paths: Set[Tuple[str, ...]] = {
        tuple([inp.get_subtree(path[:idx]).value
               for idx in range(len(path))])
        for inp in inputs
        for path, _ in inp.leaves()
    }

    input_reachability_relation: Set[Tuple[str, str]] = transitive_closure({
        pair
        for path in nonterminal_paths
        for pair in [(path[i], path[k]) for i in range(len(path)) for k in range(i + 1, len(path))]
    })

    logger.debug("Computed input reachability relation of size %d", len(input_reachability_relation))

    return input_reachability_relation


def patterns_from_file(file_name: str = STANDARD_PATTERNS_REPO) -> PatternRepository:
    if os.path.isfile(file_name):
        f = open(file_name, "r")
        contents = f.read()
        f.close()
    else:
        contents = pkgutil.get_data("islearn", STANDARD_PATTERNS_REPO).decode("UTF-8")

    data: Dict[str, List[Dict[str, str]]] = cast(Dict[str, List[Dict[str, str]]], toml.loads(contents))
    assert isinstance(data, dict)
    assert len(data) > 0
    assert all(isinstance(entry, str) for entry in data.keys())
    assert all(isinstance(entry, list) for entry in data.values())

    return PatternRepository(data)



# Generated from MexprParser.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,14,48,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,1,0,4,0,10,8,0,11,0,12,
        0,11,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,5,1,24,8,1,10,1,12,
        1,27,9,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,36,8,1,1,2,1,2,1,2,1,2,
        3,2,42,8,2,1,3,1,3,1,3,1,3,1,3,0,0,4,0,2,4,6,0,0,49,0,9,1,0,0,0,
        2,35,1,0,0,0,4,41,1,0,0,0,6,43,1,0,0,0,8,10,3,2,1,0,9,8,1,0,0,0,
        10,11,1,0,0,0,11,9,1,0,0,0,11,12,1,0,0,0,12,1,1,0,0,0,13,14,5,1,
        0,0,14,15,3,6,3,0,15,16,5,9,0,0,16,17,5,5,0,0,17,36,1,0,0,0,18,19,
        5,1,0,0,19,20,5,6,0,0,20,25,3,4,2,0,21,22,5,8,0,0,22,24,3,4,2,0,
        23,21,1,0,0,0,24,27,1,0,0,0,25,23,1,0,0,0,25,26,1,0,0,0,26,28,1,
        0,0,0,27,25,1,0,0,0,28,29,5,7,0,0,29,30,5,5,0,0,30,36,1,0,0,0,31,
        32,5,2,0,0,32,33,5,14,0,0,33,36,5,13,0,0,34,36,5,3,0,0,35,13,1,0,
        0,0,35,18,1,0,0,0,35,31,1,0,0,0,35,34,1,0,0,0,36,3,1,0,0,0,37,38,
        3,6,3,0,38,39,5,9,0,0,39,42,1,0,0,0,40,42,5,9,0,0,41,37,1,0,0,0,
        41,40,1,0,0,0,42,5,1,0,0,0,43,44,5,11,0,0,44,45,5,9,0,0,45,46,5,
        10,0,0,46,7,1,0,0,0,4,11,25,35,41
    ]

class MexprParser ( Parser ):

    grammarFileName = "MexprParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "'['", "<INVALID>", "<INVALID>", 
                     "'}'", "'<?MATCHEXPR('", "')>'", "','", "<INVALID>", 
                     "'>'", "'<'", "<INVALID>", "']'" ]

    symbolicNames = [ "<INVALID>", "BRAOP", "OPTOP", "TEXT", "NL", "BRACL", 
                      "PLACEHOLDER_OP", "PLACEHOLDER_CL", "COMMA", "ID", 
                      "GT", "LT", "WS", "OPTCL", "OPTTXT" ]

    RULE_matchExpr = 0
    RULE_matchExprElement = 1
    RULE_mexprPlaceholderParam = 2
    RULE_varType = 3

    ruleNames =  [ "matchExpr", "matchExprElement", "mexprPlaceholderParam", 
                   "varType" ]

    EOF = Token.EOF
    BRAOP=1
    OPTOP=2
    TEXT=3
    NL=4
    BRACL=5
    PLACEHOLDER_OP=6
    PLACEHOLDER_CL=7
    COMMA=8
    ID=9
    GT=10
    LT=11
    WS=12
    OPTCL=13
    OPTTXT=14

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class MatchExprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def matchExprElement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(MexprParser.MatchExprElementContext)
            else:
                return self.getTypedRuleContext(MexprParser.MatchExprElementContext,i)


        def getRuleIndex(self):
            return MexprParser.RULE_matchExpr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatchExpr" ):
                listener.enterMatchExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatchExpr" ):
                listener.exitMatchExpr(self)




    def matchExpr(self):

        localctx = MexprParser.MatchExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_matchExpr)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 9 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 8
                self.matchExprElement()
                self.state = 11 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MexprParser.BRAOP) | (1 << MexprParser.OPTOP) | (1 << MexprParser.TEXT))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MatchExprElementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return MexprParser.RULE_matchExprElement

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class MatchExprCharsContext(MatchExprElementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a MexprParser.MatchExprElementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TEXT(self):
            return self.getToken(MexprParser.TEXT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatchExprChars" ):
                listener.enterMatchExprChars(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatchExprChars" ):
                listener.exitMatchExprChars(self)


    class MatchExprOptionalContext(MatchExprElementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a MexprParser.MatchExprElementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def OPTOP(self):
            return self.getToken(MexprParser.OPTOP, 0)
        def OPTTXT(self):
            return self.getToken(MexprParser.OPTTXT, 0)
        def OPTCL(self):
            return self.getToken(MexprParser.OPTCL, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatchExprOptional" ):
                listener.enterMatchExprOptional(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatchExprOptional" ):
                listener.exitMatchExprOptional(self)


    class MatchExprVarContext(MatchExprElementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a MexprParser.MatchExprElementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def BRAOP(self):
            return self.getToken(MexprParser.BRAOP, 0)
        def varType(self):
            return self.getTypedRuleContext(MexprParser.VarTypeContext,0)

        def ID(self):
            return self.getToken(MexprParser.ID, 0)
        def BRACL(self):
            return self.getToken(MexprParser.BRACL, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatchExprVar" ):
                listener.enterMatchExprVar(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatchExprVar" ):
                listener.exitMatchExprVar(self)


    class MatchExprPlaceholderContext(MatchExprElementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a MexprParser.MatchExprElementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def BRAOP(self):
            return self.getToken(MexprParser.BRAOP, 0)
        def PLACEHOLDER_OP(self):
            return self.getToken(MexprParser.PLACEHOLDER_OP, 0)
        def mexprPlaceholderParam(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(MexprParser.MexprPlaceholderParamContext)
            else:
                return self.getTypedRuleContext(MexprParser.MexprPlaceholderParamContext,i)

        def PLACEHOLDER_CL(self):
            return self.getToken(MexprParser.PLACEHOLDER_CL, 0)
        def BRACL(self):
            return self.getToken(MexprParser.BRACL, 0)
        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(MexprParser.COMMA)
            else:
                return self.getToken(MexprParser.COMMA, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatchExprPlaceholder" ):
                listener.enterMatchExprPlaceholder(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatchExprPlaceholder" ):
                listener.exitMatchExprPlaceholder(self)



    def matchExprElement(self):

        localctx = MexprParser.MatchExprElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_matchExprElement)
        self._la = 0 # Token type
        try:
            self.state = 35
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                localctx = MexprParser.MatchExprVarContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 13
                self.match(MexprParser.BRAOP)
                self.state = 14
                self.varType()
                self.state = 15
                self.match(MexprParser.ID)
                self.state = 16
                self.match(MexprParser.BRACL)
                pass

            elif la_ == 2:
                localctx = MexprParser.MatchExprPlaceholderContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 18
                self.match(MexprParser.BRAOP)
                self.state = 19
                self.match(MexprParser.PLACEHOLDER_OP)
                self.state = 20
                self.mexprPlaceholderParam()
                self.state = 25
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==MexprParser.COMMA:
                    self.state = 21
                    self.match(MexprParser.COMMA)
                    self.state = 22
                    self.mexprPlaceholderParam()
                    self.state = 27
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 28
                self.match(MexprParser.PLACEHOLDER_CL)
                self.state = 29
                self.match(MexprParser.BRACL)
                pass

            elif la_ == 3:
                localctx = MexprParser.MatchExprOptionalContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 31
                self.match(MexprParser.OPTOP)
                self.state = 32
                self.match(MexprParser.OPTTXT)
                self.state = 33
                self.match(MexprParser.OPTCL)
                pass

            elif la_ == 4:
                localctx = MexprParser.MatchExprCharsContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 34
                self.match(MexprParser.TEXT)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MexprPlaceholderParamContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def varType(self):
            return self.getTypedRuleContext(MexprParser.VarTypeContext,0)


        def ID(self):
            return self.getToken(MexprParser.ID, 0)

        def getRuleIndex(self):
            return MexprParser.RULE_mexprPlaceholderParam

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMexprPlaceholderParam" ):
                listener.enterMexprPlaceholderParam(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMexprPlaceholderParam" ):
                listener.exitMexprPlaceholderParam(self)




    def mexprPlaceholderParam(self):

        localctx = MexprParser.MexprPlaceholderParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_mexprPlaceholderParam)
        try:
            self.state = 41
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [MexprParser.LT]:
                self.enterOuterAlt(localctx, 1)
                self.state = 37
                self.varType()
                self.state = 38
                self.match(MexprParser.ID)
                pass
            elif token in [MexprParser.ID]:
                self.enterOuterAlt(localctx, 2)
                self.state = 40
                self.match(MexprParser.ID)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VarTypeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LT(self):
            return self.getToken(MexprParser.LT, 0)

        def ID(self):
            return self.getToken(MexprParser.ID, 0)

        def GT(self):
            return self.getToken(MexprParser.GT, 0)

        def getRuleIndex(self):
            return MexprParser.RULE_varType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVarType" ):
                listener.enterVarType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVarType" ):
                listener.exitVarType(self)




    def varType(self):

        localctx = MexprParser.VarTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_varType)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 43
            self.match(MexprParser.LT)
            self.state = 44
            self.match(MexprParser.ID)
            self.state = 45
            self.match(MexprParser.GT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import logging
import os
import string
import subprocess
import sys
import tempfile
from subprocess import PIPE

import graphviz
# NOTE: To make this a PEG grammar, we need to escape single quotes within
#       multiline literal strings. So, a multiline string with """ is the
#       same as a multiline string with ''', just with different quotation marks.
from isla import language
from isla.helpers import srange, crange
from isla.type_defs import Grammar

toml_grammar = {
    "<start>": ["<document>"],
    "<document>": ["<expressions>"],
    "<expressions>": ["<expression><NL><expressions>", "<expression>"],
    "<expression>": ["<table><comment>", "<key_value><comment>", "<comment>"],
    "<comment>": ["# <STR_NO_NL>", ""],
    "<STR_NO_NL>": ["<STR_NO_NL_CHARS><STR_NO_NL>", ""],
    "<STR_NO_NL_CHARS>": list(set(srange(string.printable)) - {"\n"}),
    "<key_value>": ["<key> = <value>"],
    "<key>": ["<dotted_key>", "<simple_key>"],
    "<simple_key>": ["<quoted_key>", "<unquoted_key>"],
    "<unquoted_key>": ["<UNQUOTED_KEY>"],
    "<quoted_key>": ["<BASIC_STRING>", "<LITERAL_STRING>"],
    "<dotted_key>": ["<simple_key><dot_simple_key>"],
    "<dot_simple_key>": [".<simple_key><dot_simple_key>", ".<simple_key>"],
    "<value>": ["<string>", "<date_time>", "<LOCAL_DATE>", "<floating_point>", "<integer>", "<bool>", "<array>",
                "<inline_table>"],
    "<string>": ["<ML_BASIC_STRING>", "<BASIC_STRING>", "<ML_LITERAL_STRING>", "<LITERAL_STRING>"],
    "<integer>": ["<HEX_INT>", "<OCT_INT>", "<BIN_INT>", "<DEC_INT>"],
    "<floating_point>": ["<FLOAT>", "<INF>", "<NAN>"],
    "<bool>": ["<BOOLEAN>"],
    "<date_time>": ["<OFFSET_DATE_TIME>", "<LOCAL_DATE_TIME>", "<LOCAL_DATE>", "<LOCAL_TIME>"],
    "<array>": ["[<opt_array_values><comment_nl_or_nl>]"],
    "<opt_array_values>": ["<array_values>", ""],
    "<array_values>": [
        "<comment_nl_or_nl><value><nl_or_comment>,<array_values><comment_nl_or_nl>",
        "<comment_nl_or_nl><value><nl_or_comment><opt_comma>"
    ],
    "<opt_comma>": [",", ""],
    "<comment_nl_or_nl>": [
        "<COMMENT><NL><comment_nl_or_nl>",
        "<COMMENT><NL>",
        "<NL><comment_nl_or_nl>",
        "<NL>",
        ""],
    "<nl_or_comment>": [
        "<NL><COMMENT><nl_or_comment>",
        "<NL><nl_or_comment>",
        "<NL><COMMENT>",
        "<NL>",
        ""],
    "<table>": [
        "<array_table>",
        "<standard_table>",
    ],
    "<standard_table>": ["[<key>]"],
    "<inline_table>": ["{<inline_table_keyvals>}"],
    "<inline_table_keyvals>": ["<inline_table_keyvals_non_empty>", ""],
    "<inline_table_keyvals_non_empty>": [
        "<key> = <value>, <inline_table_keyvals_non_empty>",
        "<key> = <value>"
    ],
    "<array_table>": ["[[<key>]]"],

    "<NL>": ["\n"],
    "<COMMENT>": ["# <STR_NO_NL>"],

    "<DIGIT>": srange(string.digits),
    "<ALPHA>": srange(string.ascii_letters),

    # booleans
    "<BOOLEAN>": ["true", "false"],

    # strings
    "<ESC>": [
        '\\"', "\\\\", "\\/", "\\b", "\\f", "\\n", "\\r", "\\t", "\\<UNICODE>", "\\<EX_UNICODE>",
    ],
    "<ML_ESC>": ["<ESC>", "\\<opt_carriage_return>\n"],
    "<opt_carriage_return>": ["\r", ""],
    "<UNICODE>": ["u<HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT>"],
    "<EX_UNICODE>": ["U<HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT><HEX_DIGIT>"],
    "<BASIC_STRING>": ['"<esc_or_no_string_endings>"'],
    "<esc_or_no_string_endings>": ["<esc_or_no_string_ending><esc_or_no_string_endings>", ""],
    "<esc_or_no_string_ending>": ["<ESC>", "<no_string_ending>"],
    "<no_string_ending>": list(set(srange(string.printable)) - set(srange('"\\\n'))),
    "<ML_BASIC_STRING>": ['"""<ml_esc_or_no_string_endings>"""'],
    "<ml_esc_or_no_string_endings>": ["<ml_esc_or_no_string_ending><ml_esc_or_no_string_endings>", ""],
    "<ml_esc_or_no_string_ending>": ["<ML_ESC>", "<no_ml_string_ending>"],
    "<no_ml_string_ending>": list(set(srange(string.printable)) - set(srange('"\\'))),
    "<LITERAL_STRING>": ["'<no_literal_string_endings>'"],
    "<no_literal_string_endings>": ["<no_literal_string_ending><no_literal_string_endings>", ""],
    "<no_literal_string_ending>": list(set(srange(string.printable)) - set(srange("'\n"))),
    "<ML_LITERAL_STRING>": ["'''<ml_esc_or_no_literal_string_endings>'''"],
    "<ml_esc_or_no_literal_string_endings>": [
        "<ml_esc_or_no_literal_string_ending><ml_esc_or_no_literal_string_endings>",
        ""],
    "<ml_esc_or_no_literal_string_ending>": ["<ML_ESC>", "<no_ml_literal_string_ending>"],
    "<no_ml_literal_string_ending>": list(set(srange(string.printable)) - set(srange("'\\"))),
    # "<any_chars>": ["<any_char><any_chars>", ""],
    # "<any_char>": srange(string.printable),
    # floating point numbers
    "<EXP>": ["<e><opt_plusminus><ZERO_PREFIXABLE_INT>"],
    "<e>": ["e", "E"],
    "<opt_plusminus>": ["<plusminus>", ""],
    "<plusminus>": ["+", "-"],
    "<ZERO_PREFIXABLE_INT>": ["<DIGIT><DIGIT_OR_UNDERSCORES>"],
    "<DIGIT_OR_UNDERSCORES>": [
        "<DIGIT_OR_UNDERSCORE><DIGIT_OR_UNDERSCORES>",
        "<DIGIT_OR_UNDERSCORE>",
        ""
    ],
    "<DIGIT_OR_UNDERSCORE>": ["<DIGIT>", "_<DIGIT>"],
    "<FRAC>": [".<ZERO_PREFIXABLE_INT>"],
    "<FLOAT>": ["<DEC_INT><FRAC><EXP>", "<DEC_INT><FRAC>", "<DEC_INT><EXP>"],
    "<INF>": ["<opt_plusminus>inf"],
    "<NAN>": ["<opt_plusminus>nan"],
    # integers
    "<HEX_DIGIT>": srange("abcdefABCDEF") + ["<DIGIT>"],
    "<DIGIT_1_9>": srange("123456789"),
    "<DIGIT_0_7>": srange("01234567"),
    "<DIGIT_0_1>": ["0", "1"],
    "<DEC_INT>": [
        "<opt_plusminus><DIGIT_1_9><DIGIT_OR_UNDERSCORE><DIGIT_OR_UNDERSCORES>",
        "<opt_plusminus><DIGIT>"
    ],
    "<HEX_INT>": ["0x<HEX_DIGIT><HEX_DIGIT_OR_UNDERSCORES>"],
    "<HEX_DIGIT_OR_UNDERSCORES>": [
        "<HEX_DIGIT_OR_UNDERSCORE><HEX_DIGIT_OR_UNDERSCORES>",
        "<HEX_DIGIT_OR_UNDERSCORE>",
        ""
    ],
    "<HEX_DIGIT_OR_UNDERSCORE>": ["_<HEX_DIGIT>", "<HEX_DIGIT>"],
    "<OCT_INT>": ["0o<DIGIT_0_7><DIGIT_0_7_OR_UNDERSCORES>"],
    "<DIGIT_0_7_OR_UNDERSCORES>": [
        "<DIGIT_0_7_OR_UNDERSCORE><DIGIT_0_7_OR_UNDERSCORES>",
        "<DIGIT_0_7_OR_UNDERSCORE>",
        ""
    ],
    "<DIGIT_0_7_OR_UNDERSCORE>": ["<DIGIT_0_7>", "_<DIGIT_0_7>"],
    "<BIN_INT>": ["0b<DIGIT_0_1><DIGIT_0_1_OR_UNDERSCORES>"],
    "<DIGIT_0_1_OR_UNDERSCORES>": [
        "<DIGIT_0_1_OR_UNDERSCORE><DIGIT_0_1_OR_UNDERSCORES>",
        "<DIGIT_0_1_OR_UNDERSCORE>",
        ""
    ],
    "<DIGIT_0_1_OR_UNDERSCORE>": [
        "_<DIGIT_0_1>",
        "<DIGIT_0_1>",
    ],
    # dates
    "<YEAR>": ["<DIGIT><DIGIT><DIGIT><DIGIT>"],
    "<MONTH>": ["<DIGIT><DIGIT>"],
    "<DAY>": ["<DIGIT><DIGIT>"],
    "<DELIM>": ["T", "t", " "],
    "<HOUR>": ["<DIGIT><DIGIT>"],
    "<MINUTE>": ["<DIGIT><DIGIT>"],
    "<SECOND>": ["<DIGIT><DIGIT>"],
    "<SECFRAC>": [".<DIGITS>"],
    "<DIGITS>": ["<DIGIT>", "<DIGIT><DIGITS>"],
    "<NUMOFFSET>": ["<plusminus><HOUR>:<MINUTE>"],
    "<OFFSET>": ["Z", "<NUMOFFSET>"],
    "<PARTIAL_TIME>": [
        "<HOUR>:<MINUTE>:<SECOND><SECFRAC>",
        "<HOUR>:<MINUTE>:<SECOND>",
    ],
    "<FULL_DATE>": ["<YEAR>-<MONTH>-<DAY>"],
    "<FULL_TIME>": ["<PARTIAL_TIME><OFFSET>"],
    "<OFFSET_DATE_TIME>": ["<FULL_DATE><DELIM><FULL_TIME>"],
    "<LOCAL_DATE_TIME>": ["<FULL_DATE><DELIM><PARTIAL_TIME>"],
    "<LOCAL_DATE>": ["<FULL_DATE>"],
    "<LOCAL_TIME>": ["<PARTIAL_TIME>"],
    # keys
    "<UNQUOTED_KEY>": ["<UNQUOTED_KEY_CHARS>"],
    "<UNQUOTED_KEY_CHARS>": ["<UNQUOTED_KEY_CHAR><UNQUOTED_KEY_CHARS>", "<UNQUOTED_KEY_CHAR>"],
    "<UNQUOTED_KEY_CHAR>": ["<ALPHA>", "<DIGIT>", "-", "_"]
}

CHARACTERS_WITHOUT_QUOTE = (
        string.digits
        + string.ascii_letters
        + string.punctuation.replace('"', '').replace('\\', '')
        + ' ')

JSON_GRAMMAR: Grammar = {
    "<start>": ["<json>"],
    "<json>": ["<element>"],
    "<element>": ["<ws><value><ws>"],
    "<value>": ["<object>", "<array>", "<string>", "<number>", "true", "false", "null"],
    "<object>": ["{<members>}", "{<ws>}"],
    "<members>": ["<member>,<members>", "<member>"],
    "<member>": ["<ws><string><ws>:<element>"],
    "<array>": ["[<ws>]", "[<elements>]"],
    "<elements>": ["<element>,<elements>", "<element>"],
    "<string>": ['"' + "<characters>" + '"'],
    "<characters>": ["<character><characters>", ""],
    "<character>": srange(CHARACTERS_WITHOUT_QUOTE),
    "<number>": ["<int><frac><exp>"],
    "<int>": ["-<onenine><digits>", "<onenine><digits>", "-<digit>", "<digit>"],
    "<digits>": ["<digit><digits>", "<digit>"],
    "<digit>": ['0', "<onenine>"],
    "<onenine>": ["1", "2", "3", "4", "5", "6", "7", "8", "9"],
    "<frac>": [".<digits>", ""],
    "<exp>": ["E<sign><digits>", "e<sign><digits>", ""],
    "<sign>": ['+', '-', ""],
    "<ws>": [" "]
}

ICMP_GRAMMAR = {
    "<start>": ["<icmp_message>"],
    "<icmp_message>": ["<header><payload_data>"],
    "<header>": ["<type><code><checksum><header_data>"],
    "<payload_data>": ["<bytes>"],
    "<type>": ["<byte>"],
    "<code>": ["<byte>"],
    "<checksum>": ["<byte><byte>"],
    "<header_data>": ["<byte><byte><byte><byte>"],
    "<byte>": ["<zerof><zerof> "],
    "<bytes>": ["<byte><bytes>", ""],
    "<zerof>": srange(string.digits + "ABCDEF")
}

# https://en.wikipedia.org/wiki/IPv4#Packet_structure
IPv4_GRAMMAR = {
    "<start>": ["<ip_message>"],
    "<ip_message>": ["<header><data>"],
    "<header>": [  # Each line 4 bytes
        "<version_ihl><dscp_ecn><total_length>"
        "<identification><flags_offset>"
        "<ttl><protocol><header_checksum>"
        "<source_ip>"
        "<dest_ip>"
    ],
    "<data>": ["<bytes>"],
    "<version_ihl>": ["<byte>"],
    "<dscp_ecn>": ["<byte>"],
    "<total_length>": ["<byte><byte>"],
    "<identification>": ["<byte><byte>"],
    "<flags_offset>": ["<byte><byte>"],
    "<ttl>": ["<byte>"],
    "<protocol>": ["<byte>"],
    "<header_checksum>": ["<byte><byte>"],
    "<source_ip>": ["<byte><byte><byte><byte>"],
    "<dest_ip>": ["<byte><byte><byte><byte>"],
    "<byte>": ["<zerof><zerof> "],
    "<bytes>": ["<byte><bytes>", ""],
    "<zerof>": srange(string.digits + "ABCDEF")
}

# Grammar source: `https://github.com/antlr/grammars-v4/blob/master/dot/DOT.g4`
# We changed the DIGRAPH, GRAPH, NODE, EDGE, SUBGRAPH, and STRICT identifiers to be case sensitive
# (lower case) to enable learning String equations. In all GitHub examples we found /
# use, this is anyway the case.
DOT_GRAMMAR = {
    "<start>": ["<graph>"],
    "<graph>": [
        "<maybe_strict><graph_type><maybe_space_id><MWSS>{<MWSS><stmt_list><MWSS>}",
    ],
    "<graph_type>": ["<GRAPH>", "<DIGRAPH>"],
    "<stmt_list>": ["<stmt><MWSS><maybe_semi><MWSS><stmt_list>", "<stmt><MWSS><maybe_semi>", ""],
    "<stmt>": ["<edge_stmt>", "<attr_stmt>", "<id><MWSS>=<MWSS><id>", "<subgraph>", "<node_stmt>"],
    "<attr_stmt>": ["<GRAPH><MWSS><attr_list>", "<NODE><MWSS><attr_list>", "<EDGE><MWSS><attr_list>"],
    "<maybe_attr_list>": ["<attr_list>", ""],
    "<attr_list>": ["[<MWSS><a_list><MWSS>]<MWSS><attr_list>", "[<MWSS><a_list><MWSS>]"],
    "<a_list>": ["<id><MWSS>=<MWSS><id><MWSS><maybe_comma><MWSS><a_list>", "<id><MWSS>=<MWSS><id><MWSS><maybe_comma>"],
    "<edge_stmt>": [
        "<node_id><MWSS><edgeRHS><MWSS><maybe_attr_list>",
        "<subgraph><MWSS><edgeRHS><MWSS><maybe_attr_list>",
    ],
    "<edgeRHS>": [
        "<edgeop><MWSS><node_id><MWSS><edgeRHS>",
        "<edgeop><MWSS><subgraph><MWSS><edgeRHS>",
        "<edgeop><MWSS><node_id>",
        "<edgeop><MWSS><subgraph>",
    ],
    "<edgeop>": ["->", "--"],
    "<node_stmt>": ["<node_id><MWSS><maybe_attr_list>"],
    "<node_id>": ["<id><MWSS><maybe_port>"],
    "<maybe_port>": ["<port>", ""],
    "<port>": [
        ":<MWSS><id><MWSS>:<MWSS><id>",
        ":<MWSS><id>",
    ],
    "<subgraph>": [
        "<SUBGRAPH><MWSS><maybe_space_id><MWSS>{<MWSS><stmt_list><MWSS>}",
        "{<MWSS><stmt_list><MWSS>}",
    ],
    "<maybe_space_id>": ["<WSS><id>", ""],
    "<id>": ["<STRING>", "<ID>", "<NUMBER>"],
    "<maybe_strict>": ["<STRICT><WSS>", ""],
    "<STRICT>": ["strict"],  # ["<S><T><R><I><C><T>"],
    "<GRAPH>": ["graph"],  # ["<G><R><A><P><H>"],
    "<DIGRAPH>": ["digraph"],  # ["<D><I><G><R><A><P><H>"],
    "<NODE>": ["node"],  # ["<N><O><D><E>"],
    "<EDGE>": ["edge"],  # ["<E><D><G><E>"],
    "<SUBGRAPH>": ["subgraph"],  # ["<S><U><B><G><R><A><P><H>"],
    "<NUMBER>": [
        "<maybe_minus><DIGITS>.<MAYBE_DIGITS>",
        "<maybe_minus><DIGITS>",
        "<maybe_minus>.<DIGITS>",
    ],
    "<MAYBE_DIGITS>": ["<DIGITS>", ""],
    "<DIGITS>": ["<DIGIT><DIGITS>", "<DIGIT>"],
    "<DIGIT>": srange(string.digits),

    "<STRING>": ['"<esc_or_no_string_endings>"'],
    "<esc_or_no_string_endings>": ["<esc_or_no_string_ending><esc_or_no_string_endings>", ""],
    "<esc_or_no_string_ending>": ['\\"', "<no_string_ending>"],
    "<no_string_ending>": list(set(srange(string.printable)) - set(srange('"\\'))),

    "<ID>": ["<LETTER><LETTER_OR_DIGITS>", ],
    "<LETTER_OR_DIGITS>": [
        "<LETTER><LETTER_OR_DIGITS>",
        "<DIGIT><LETTER_OR_DIGITS>",
        ""
    ],
    "<LETTER>": srange(string.ascii_letters + "_"),
    "<maybe_minus>": ["-", ""],
    "<maybe_comma>": [",", ""],
    "<maybe_semi>": [";", ""],
    # "<A>": ["A", "a"],
    # "<B>": ["B", "b"],
    # "<C>": ["C", "c"],
    # "<D>": ["D", "d"],
    # "<E>": ["E", "e"],
    # "<G>": ["G", "g"],
    # "<H>": ["H", "h"],
    # "<I>": ["I", "i"],
    # "<N>": ["N", "n"],
    # "<O>": ["O", "o"],
    # "<P>": ["P", "p"],
    # "<R>": ["R", "r"],
    # "<S>": ["S", "s"],
    # "<T>": ["T", "t"],
    # "<U>": ["U", "u"],
    "<MWSS>": ["<WSS>", ""],
    "<WSS>": ["<WS><WSS>", "<WS>"],
    "<WS>": [" ", "\t", "\n"]
}


def render_dot(inp: language.DerivationTree | str) -> bool | str:
    logging.getLogger("graphviz.backend").propagate = False
    logging.getLogger("graphviz.files").propagate = False

    with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
        devnull = open(os.devnull, 'w')
        orig_stderr = sys.stderr
        sys.stderr = devnull

        err_message = ""
        try:
            graphviz.Source(str(inp) if isinstance(inp, language.DerivationTree) else inp).render(outfile.name)
        except Exception as exc:
            err_message = str(exc)

        sys.stderr = orig_stderr
        return True if not err_message else err_message


# Grammar source: `https://github.com/antlr/grammars-v4/blob/master/racket-bsl/BSL.g4`
RACKET_BSL_GRAMMAR = {
    "<start>": ["<program>"],
    "<program>": ["<def_or_exprs>"],
    "<def_or_exprs>": ["<def_or_expr><MWSS><def_or_exprs>", "<def_or_expr>"],
    "<def_or_expr>": [
        "<maybe_comments><definition>",
        "<expr>",
        "<test_case>",
        "<library_require>",
        "<HASHDIRECTIVE>",
        "<COMMENT>"
    ],

    "<definition>": [
        "(<MWSS>define<MWSS>(<MWSS><name><WSS_NAMES><MWSS>)<MWSS><expr><MWSS>)",
        "(<MWSS>define<WSS><name><MWSS><expr><MWSS>)",
        "(<MWSS>define<WSS><name><MWSS>(lambda<MWSS>(<MWSS><WSS_NAMES><MWSS>)<MWSS><expr><MWSS>)<MWSS>)",
        "(<MWSS>define-struct<WSS><name><MWSS>(<MWSS><name><maybe_wss_names><MWSS>)<MWSS>)",
        "(<MWSS>define-struct<WSS><name><MWSS>(<MWSS>)<MWSS>)",
    ],

    "<WSS_NAMES>": ["<WSS><NAME><WSS_NAMES>", "<WSS><NAME>"],
    "<maybe_wss_names>": ["<WSS_NAMES>", ""],

    "<wss_exprs>": ["<MWSS><expr><wss_exprs>", "<MWSS><expr>"],

    "<cond_args>": [
        "<maybe_comments><MWSS>[<MWSS><expr><WSS><expr><MWSS>]<MWSS><cond_args>",
        "<maybe_comments><MWSS>[<MWSS><expr><WSS><expr><MWSS>]"
    ],

    "<maybe_cond_args>": ["<cond_args>", ""],

    "<expr>": [
        "<maybe_comments><MWSS>(<MWSS>cond<MWSS><maybe_cond_args><MWSS>[else<WSS><expr>]<MWSS>)",
        "<maybe_comments><MWSS>(<MWSS>cond<MWSS><cond_args><MWSS>)",
        "<maybe_comments><MWSS>(<MWSS>if<WSS><expr><WSS><expr><WSS><expr><MWSS>)",
        "<maybe_comments><MWSS>(<MWSS>and<WSS><expr><wss_exprs><MWSS>)",
        "<maybe_comments><MWSS>(<MWSS>or<WSS><expr><wss_exprs><MWSS>)",
        "<maybe_comments><MWSS>(<MWSS><name><wss_exprs><MWSS>)",
        "<maybe_comments><MWSS>'()",
        "<maybe_comments><MWSS><STRING>",
        "<maybe_comments><MWSS><NUMBER>",
        "<maybe_comments><MWSS><BOOLEAN>",
        "<maybe_comments><MWSS><CHARACTER>",
        "<maybe_comments><MWSS><name>",
    ],

    "<test_case>": [
        "(<MWSS>check-expect<WSS><expr><WSS><expr><MWSS>)",
        "(<MWSS>check-random<WSS><expr><WSS><expr><MWSS>)",
        "(<MWSS>check-within<WSS><expr><WSS><expr><WSS><expr><MWSS>)",
        "(<MWSS>check-member-of<WSS><expr><wss_exprs><MWSS>)",
        "(<MWSS>check-satisfied<WSS><expr><WSS><name><MWSS>)",
        "(<MWSS>check-error<WSS><expr><WSS><expr><MWSS>)",
        "(<MWSS>check-error<WSS><expr><MWSS>)",
    ],

    "<library_require>": [
        "(<MWSS>require<MWSS><STRING><MWSS>)",
        "(<MWSS>require<WSS><name><MWSS>)",
        "(<MWSS>require<MWSS>(<MWSS><name><MWSS><STRING><MWSS>(<MWSS><strings_mwss>)<MWSS>)<MWSS>)",
        "(<MWSS>require<MWSS>(<MWSS><name><MWSS><STRING><MWSS>)<MWSS>)",
        "(<MWSS>require<MWSS>(<MWSS><name><MWSS><STRING><MWSS><pkg><MWSS>)<MWSS>)",
    ],

    "<strings_mwss>": ["<STRING><MWSS><strings_mwss>", "<STRING>"],

    "<pkg>": ["(<MWSS><STRING><MWSS><STRING><WSS><NUMBER><WSS><NUMBER><MWSS>)"],

    "<name>": ["<SYMBOL>", "<NAME>"],

    #  A symbol is a quote character followed by a name. A symbol is a value, just like 42, '(), or #false.
    "<SYMBOL>": ["'<NAME>"],

    # A name or a variable is a sequence of characters not including a space or one of the following:
    #   " , ' ` ( ) [ ] { } | ; #
    # Original expression: ([$%&!*+\\^_~]|[--:<-Za-z])+
    "<NAME>": ["<NAME_CHARS>"],
    "<NAME_CHARS>": ["<NAME_CHAR><NAME_CHARS>", "<NAME_CHAR>"],
    "<NAME_CHAR>": srange("$%&!*+\\^_~") + crange("-", ":") + crange("<", "Z") + crange("a", "z"),

    # A number is a number such as 123, 3/2, or 5.5.
    "<NUMBER>": [
        "<INT>.<DIGITS>",
        "<INT>/<INT>",
        "INT",
    ],

    "<DIGITS>": ["<DIGIT><DIGITS>", "<DIGIT>"],
    "<MAYBE_DIGITS>": ["<DIGITS>", ""],
    "<ONENINE>": srange("123456789"),

    "<INT>": ["<ONENINE><MAYBE_DIGITS>", "0"],

    "<BOOLEAN>": ["#true", "#T", "#t", "#false", "#F", "#f", ],

    # A string is a sequence of characters enclosed by a pair of ".
    # Unlike symbols, strings may be split into characters and manipulated by a variety of functions.
    # For example, "abcdef", "This is a string", and "This is a string with \" inside" are all strings.
    "<STRING>": ['"<ESC_OR_NO_STRING_ENDINGS>"'],
    "<ESC_OR_NO_STRING_ENDINGS>": ["<ESC_OR_NO_STRING_ENDING><ESC_OR_NO_STRING_ENDINGS>", ""],
    "<ESC_OR_NO_STRING_ENDING>": ['\\"', "<NO_STRING_ENDING>"],
    "<NO_STRING_ENDING>": list(set(srange(string.printable)) - set(srange('"\\'))),

    # A character begins with #\ and has the name of the character.
    # For example, #\a, #\b, and #\space are characters.
    "<CHARACTER>": ["#\\space", "#\\<LETTERORDIGIT>"],

    "<DIGIT>": srange(string.digits),
    "<LETTERORDIGIT>": srange(string.ascii_letters + string.digits),

    "<MWSS>": ["<WSS>", ""],
    "<WSS>": ["<WS><WSS>", "<WS>"],
    "<WS>": [" ", "\t", "\n"],

    "<maybe_comments>": ["<COMMENT><MWSS><maybe_comments>", ""],
    "<COMMENT>": [";<NOBRs>\n"],
    "<HASHDIRECTIVE>": ["#lang <NOBRs>", "#reader<NOBRs>"],
    "<NOBR>": list(set(string.printable) - {"\n", "\r"}),
    "<NOBRs>": ["<NOBR><NOBRs>", "<NOBR>"],
}


def load_racket(tree: language.DerivationTree | str) -> bool | str:
    with tempfile.NamedTemporaryFile(suffix=".rk") as tmp:
        tmp.write(str(tree).encode())
        tmp.flush()
        # cmd = ["racket", "-e", f'(require drracket/check-syntax) (show-content "{tmp.name}")']
        cmd = ["racket", "-f", tmp.name]
        process = subprocess.Popen(cmd, stderr=PIPE, stdin=PIPE)
        (stdout, stderr) = process.communicate()
        exit_code = process.wait()

        err_msg = stderr.decode("utf-8")
        has_error = exit_code != 0 or (bool(err_msg) and "read-syntax" in err_msg)

        # if has_error:
        #     print("===== ERROR =====")
        #     print("Program:")
        #     print(tree)
        #     print()
        #     print("Message:")
        #     print(err_msg)
        #     print("=================")

        return True if not has_error else err_msg



import logging
import random
from typing import Set, Dict, Callable, Tuple, Iterable, Generator, Optional

from grammar_graph import gg
from isla.existential_helpers import paths_between, path_to_tree
from isla.fuzzer import GrammarCoverageFuzzer
from isla.helpers import is_nonterminal, canonical
from isla.language import DerivationTree
from isla.type_defs import Grammar, Path

random = random.SystemRandom()

class MutationFuzzer:
    """
    A k-path coverage guided, grammar-aware mutation fuzzer using input fragments.
    """

    def __init__(
            self,
            grammar: Grammar,
            seed: Iterable[DerivationTree],
            property: Callable[[DerivationTree], bool] = lambda tree: True,
            k: int = 3,
            min_mutations: int = 2,
            max_mutations: int = 10):
        self.logger = logging.getLogger(type(self).__name__)
        self.grammar = grammar
        self.graph = gg.GrammarGraph.from_grammar(grammar)
        self.seed = set(seed)
        self.property = property
        self.k = k
        self.min_mutations = min_mutations
        self.max_mutations = max_mutations

        self.fuzzer = GrammarCoverageFuzzer(grammar)

        self.population = set([])
        self.coverages_seen: Set[Tuple[gg.Node, ...]] = set([])
        self.fragments: Dict[str, Set[DerivationTree]] = {}

        self.reset()

        self.mutators: Dict[Callable[[DerivationTree, Path], DerivationTree | None]] = {
            self.replace_with_fragment: 4,
            self.swap_subtrees: 3,
            self.replace_with_random_subtree: 2,
            self.generalize: 1,
        }

    def reset(self):
        self.population = set(self.seed)
        self.coverages_seen = set([])
        self.coverages_seen = {p for inp in self.population for p in self.coverages_of(inp)}

        for sample in self.population:
            self.update_fragments(sample)

    def update_fragments(self, sample: DerivationTree) -> None:
        for _, tree in sample.paths():
            self.fragments.setdefault(tree.value, set([])).add(tree)

    def coverages_of(self, inp: DerivationTree) -> Set[Tuple[gg.Node, ...]]:
        return self.graph.k_paths_in_tree(inp.to_parse_tree(), self.k)

    def fuzz(self) -> DerivationTree:
        num_mutations = random.randint(self.max_mutations, self.max_mutations)
        curr_inp = random.choice(list(self.population))
        mutations = 0
        while mutations < num_mutations:
            maybe_result = self.mutate(curr_inp)
            if maybe_result is not None:
                curr_inp = maybe_result
                mutations += 1
        return curr_inp

    def mutate(self, inp: DerivationTree) -> DerivationTree | None:
        paths = [path for path, subtree in inp.paths() if is_nonterminal(subtree.value)]
        while paths:
            path = random.choice(paths)
            paths.remove(path)

            mutators = dict(self.mutators)
            while mutators:
                mutator = random.choices(
                    list(mutators.keys()),
                    list(mutators.values()), k=1)[0]
                del mutators[mutator]

                result = mutator(inp, path)
                if result is not None:
                    return result

        return None

    def replace_with_fragment(self, inp: DerivationTree, path: Path) -> DerivationTree | None:
        subtree = inp.get_subtree(path)
        if not is_nonterminal(subtree.value):
            return None

        different_fragments = [
            fragment for fragment in self.fragments.get(subtree.value, [])
            if fragment and not fragment.structurally_equal(subtree)
        ]

        if not different_fragments:
            return inp

        result = inp.replace_path(path, random.choice(different_fragments))
        assert result is None or self.graph.tree_is_valid(result.to_parse_tree())
        return result

    def replace_with_random_subtree(self, inp: DerivationTree, path: Path) -> DerivationTree | None:
        subtree = inp.get_subtree(path)
        if not is_nonterminal(subtree.value):
            return None

        result = inp.replace_path(path, self.fuzzer.expand_tree(DerivationTree(subtree.value, None)))
        assert result is None or self.graph.tree_is_valid(result.to_parse_tree())
        return result

    def swap_subtrees(self, inp: DerivationTree, path: Path) -> DerivationTree | None:
        subtree = inp.get_subtree(path)
        if not is_nonterminal(subtree.value):
            return None

        matches = [tree for other_path, tree in inp.filter(lambda t: t.value == subtree.value)
                   if other_path != path]

        if not matches:
            return None

        result = inp.replace_path(path, random.choice(matches))
        assert result is None or self.graph.tree_is_valid(result.to_parse_tree())
        return result

    def generalize(self, inp: DerivationTree, path: Path) -> DerivationTree | None:
        subtree = inp.get_subtree(path)
        if not is_nonterminal(subtree.value):
            return None

        self_embedding_trees = [
            self_embedding_tree
            for self_embedding_path in paths_between(self.graph, subtree.value, subtree.value)
            for self_embedding_tree in path_to_tree(canonical(self.grammar), self_embedding_path)]

        if not self_embedding_trees:
            return None

        self_embedding_tree = random.choice(self_embedding_trees)
        matching_leaf = random.choice([p for p, t in self_embedding_tree.leaves() if t.value == subtree.value])

        result = inp.replace_path(path, self_embedding_tree.replace_path(matching_leaf, subtree))
        result = self.expand_tree(result)
        assert result is None or self.graph.tree_is_valid(result.to_parse_tree())

        return result

    def expand_tree(self, inp: DerivationTree) -> DerivationTree:
        for path, tree in inp.open_leaves():
            if random.random() < .3 or tree.value not in self.fragments:
                inp = inp.replace_path(path, self.fuzzer.expand_tree(tree))
            else:
                inp = inp.replace_path(path, random.choice(list(self.fragments[tree.value])))

        assert inp.is_complete()
        return inp

    def run(
            self,
            num_iterations: Optional[int] = 500,
            alpha: float = 0.1,
            extend_fragments: bool = True,
            yield_negative=False) -> Generator[DerivationTree, None, None]:
        unsuccessful_tries = 0

        i = 0
        while True:
            if num_iterations is not None and i >= num_iterations:
                break

            curr_alpha = 1 - (unsuccessful_tries / (i + 1))
            if curr_alpha < alpha:
                if i * 10 > (num_iterations or 500):
                    break

            inp = self.fuzz()

            if self.process_new_input(inp, extend_fragments):
                yield inp
            else:
                unsuccessful_tries += 1
                if yield_negative:
                    yield inp
                self.logger.debug("current alpha: %f, threshold: %f", curr_alpha, alpha)

            i += 1

    def process_new_input(self, inp: DerivationTree, extend_fragments: bool = True) -> bool:
        new_coverage = self.coverages_seen - self.coverages_of(inp)
        if inp in self.population or not self.property(inp) or not new_coverage:
            return False

        self.coverages_seen.update(new_coverage)
        self.population.add(inp)
        if extend_fragments:
            self.update_fragments(inp)

        return True
